from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class BaseOptoControlContext:
    optocontrol_key: str
    alias: str


@dataclass(frozen=True)
class MaskContext(BaseOptoControlContext):
    mask: Any
    daq_port: int
    daq_line: int




from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any

from PyQt6.QtWidgets import QWidget

from pyrpoc.backend_utils.opto_control_contexts import BaseOptoControlContext
from pyrpoc.backend_utils.state_helpers import export_object_state, import_object_state, make_instance_id


class BaseOptoControlWidget(QWidget):
    """Base class for all optocontrol UI widgets.

    This lets the optocontrol manager mount a concrete editor widget without importing
    the control subtype. Concrete widgets should call `on_change` whenever user-editable
    values change so service/state handlers can persist updates.
    """

    def __init__(
        self,
        control: "BaseOptoControl",
        on_change: Callable[[], None] | None = None,
        parent: QWidget | None = None,
    ):
        super().__init__(parent)
        self.control = control
        self.on_change = on_change

    def refresh_from_model(self) -> None:
        """Load current control data into widget fields.

        The widget manager calls this when a control is created from session restore.
        """
        return None

    def request_persist(self) -> None:
        if self.on_change is not None:
            self.on_change()


class BaseOptoControl(ABC):
    OPTOCONTROL_KEY: str = "base_optocontrol"
    DISPLAY_NAME: str = "Base Opto-Control"
    PERSISTENCE_FIELDS: tuple[str, ...] | None = None
    PERSISTENCE_EXCLUDE_FIELDS: tuple[str, ...] = (
        "alias",
        "connected",
        "enabled",
        "instance_id",
        "last_error",
        "user_label",
        "widget",
    )

    def __init__(
        self,
        alias: str | None = None,
        user_label: str | None = None,
        enabled: bool = False,
        *,
        instance_id: str | None = None,
        connected: bool = False,
    ):
        self.alias = alias or self.OPTOCONTROL_KEY
        self.instance_id = instance_id or make_instance_id(self.alias)
        self.user_label = user_label
        self.enabled = enabled
        self.connected = bool(connected)
        self.last_error: str | None = None
        self.context: BaseOptoControlContext | None = None

    @property
    def type_key(self) -> str:
        '''Return the registry key used to recreate this control on restore.'''
        return self.alias

    @classmethod
    def get_contract(cls) -> dict[str, Any]:
        return {
            "optocontrol_key": cls.OPTOCONTROL_KEY,
            "display_name": cls.DISPLAY_NAME,
        }

    @abstractmethod
    def get_widget(
        self,
        parent: QWidget | None = None,
        on_change: Callable[[], None] | None = None,
        display_service: Any | None = None,
    ) -> BaseOptoControlWidget:
        """Create or reuse the editor widget for this specific control.

        Called by OptoControlManager UI handlers when it builds/refreshes the control list.
        """
        raise NotImplementedError

    def get_summary(self) -> str:
        """Return collapsed-row text for the control list view.

        The manager displays this in the compact control card and updates it after edits.
        """
        return self.user_label or self.DISPLAY_NAME

    def get_context(self) -> BaseOptoControlContext:
        """Build the control-specific acquisition context for the current run.

        Subclasses override this to pass modality-facing, acquisition-only state.
        """
        return BaseOptoControlContext(optocontrol_key=self.OPTOCONTROL_KEY, alias=self.alias)

    def cleanup(self) -> None:
        """Release widget/state resources before removal.

        Concrete controls override this when they hold threads, file handles, or external state.
        """
        self.last_error = None

    def prepare_for_acquisition(self) -> BaseOptoControlContext:
        """Default acquisition handoff for modality execution.

        This keeps `ModalityService` and existing callers simple by returning one
        context object per control rather than arbitrary payload tuples.
        """
        self.context = self.get_context()
        return self.context

    def export_persistence_state(self) -> dict[str, Any]:
        return export_object_state(
            self,
            include_fields=self.PERSISTENCE_FIELDS,
            exclude_fields=self.PERSISTENCE_EXCLUDE_FIELDS,
        )

    def import_persistence_state(self, state: dict[str, Any]) -> None:
        import_object_state(
            self,
            state,
            include_fields=self.PERSISTENCE_FIELDS,
            exclude_fields=self.PERSISTENCE_EXCLUDE_FIELDS,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from PyQt6.QtWidgets import QMessageBox

from pyrpoc.optocontrols.base_optocontrol import BaseOptoControl
from pyrpoc.gui.main_widgets.opto_control_mgr.instance_card import InstanceCardWidget

if TYPE_CHECKING:
    from pyrpoc.gui.main_widgets.opto_control_mgr.widget import OptoControlManagerWidget


def refresh_available(widget: OptoControlManagerWidget) -> None:
    '''Read selected modality compatibility and refresh dropdown entries only for those.

    This is called by the manager on startup and whenever modality changes.
    '''
    contract = widget.modality_service.get_selected_contract()
    allowed = contract.get("allowed_optocontrols", [])
    allowed_keys = {
        getattr(opto_cls, "OPTOCONTROL_KEY", "")
        for opto_cls in allowed
        if hasattr(opto_cls, "OPTOCONTROL_KEY")
    }

    current_key = widget._selected_type_key()
    widget.type_combo.clear()
    for row in widget.opto_control_service.list_available():
        key = row["key"]
        if allowed_keys and key not in allowed_keys:
            continue
        name = row.get("display_name", key)
        widget.type_combo.addItem(name, key)

    if current_key:
        idx = widget.type_combo.findData(current_key)
        if idx >= 0:
            widget.type_combo.setCurrentIndex(idx)


def refresh_instances(widget: OptoControlManagerWidget) -> None:
    '''
    Keep card widgets stable across inventory deltas.

    Route:
    - service emits `inventory_changed` after add/remove
    - this function compares service rows with existing cards and only adds/removes
      changed control instances, preserving existing expanded bodies and QObject state.
    '''
    desired_cards: dict[BaseOptoControl, InstanceCardWidget] = {}
    rows = widget.opto_control_service.list_instances()
    wanted_controls = [row["state"] for row in rows]

    _remove_missing_cards(widget, set(wanted_controls))

    for row in rows:
        control: BaseOptoControl = row["state"]
        title = row["name"]
        card = widget.state.card_widgets.get(control)
        if card is None:
            card = _create_card(widget, control, title, row["key"])
            card.set_enable_checked(row.get("enabled", False))
            card.set_enable_visible(True)
            card.set_marker_text(f"[{row['key']}]")
            widget.state.card_widgets[control] = card

        card.set_enable_checked(row.get("enabled", False))
        _refresh_card_text(card, title, row["key"])
        desired_cards[control] = card

    _reorder_cards(widget, desired_cards, rows)


def on_expand_requested(widget: OptoControlManagerWidget, state_obj: object) -> None:
    '''Toggle the card body and lazy-build the control widget when expanding.

    The body is created here so collapsed cards remain lightweight.
    '''
    if not isinstance(state_obj, BaseOptoControl):
        return
    card = widget.state.card_widgets.get(state_obj)
    if card is None:
        return

    card.set_expanded(not card.is_expanded())
    if not card.is_expanded():
        return

    if card.body_layout.count() != 0:
        if _is_stale_widget(card):
            card.set_body_widget(None)
            _reset_control_widget_cache(state_obj)
        else:
            return

    if _attach_widget_to_card(widget, state_obj, card):
        return

    # Retry path for stale-object style failures in case the cached control widget
    # wrapper was deleted after a prior layout cleanup.
    _reset_control_widget_cache(state_obj)
    _attach_widget_to_card(widget, state_obj, card)


def on_add_clicked(widget: OptoControlManagerWidget) -> None:
    '''Add the selected type into app state and trigger a list refresh.

    Route: Add button -> this handler -> service.create_opto_control -> inventory_changed signal.
    '''
    key = widget._selected_type_key()
    if not key:
        show_error(widget, "No opto-control available for current modality")
        return
    try:
        widget.opto_control_service.create_opto_control(key)
    except Exception as exc:
        show_error(widget, str(exc))


def on_enable_toggled(widget: OptoControlManagerWidget, state_obj: object, checked: bool) -> None:
    '''Persist checkbox state directly into the control instance.

    Route: card checkbox -> this handler -> service.set_enabled.
    '''
    if not isinstance(state_obj, BaseOptoControl):
        return
    widget.opto_control_service.set_enabled(state_obj, checked)


def on_remove_requested(widget: OptoControlManagerWidget, state_obj: object) -> None:
    '''Remove control instance from state and UI.

    Route: card remove button -> this handler -> service.remove_opto_control.
    '''
    if not isinstance(state_obj, BaseOptoControl):
        return
    widget.opto_control_service.remove_opto_control(state_obj)


def on_modality_selected(widget: OptoControlManagerWidget, key: str) -> None:
    del key
    refresh_available(widget)


def show_error(widget: OptoControlManagerWidget, message: str) -> None:
    QMessageBox.critical(widget, "Opto-Control Error", message)


def _remove_missing_cards(widget: OptoControlManagerWidget, wanted_controls: set[BaseOptoControl]) -> None:
    '''
    Remove only cards whose control state is no longer present.

    This path is triggered from inventory refresh after service remove/add.
    Reusable cards are retained so expanded bodies are not force-destroyed.
    '''
    for control, card in list(widget.state.card_widgets.items()):
        if control not in wanted_controls:
            widget.state.card_widgets.pop(control)
            card.setParent(None)
            card.deleteLater()


def _create_card(
    widget: OptoControlManagerWidget,
    control: BaseOptoControl,
    title: str,
    key: str,
) -> InstanceCardWidget:
    '''
    Create one opto-control card for a newly-added control state.

    From:
    - add/remove refresh cycle (from service signal) -> this branch.
    To:
    - card instance owned by this manager so existing references map state->widget.
    '''
    card = InstanceCardWidget(control, title, widget)
    card.set_enable_visible(True)
    card.set_marker_text(f"[{key}]")
    card.remove_requested.connect(lambda state_obj, w=widget: on_remove_requested(w, state_obj))
    card.expand_requested.connect(lambda state_obj, w=widget: on_expand_requested(w, state_obj))
    card.enable_toggled.connect(
        lambda state_obj, checked, w=widget: on_enable_toggled(w, state_obj, checked)
    )
    widget.instances_layout.insertWidget(widget.instances_layout.count(), card)
    return card


def _refresh_card_text(card: InstanceCardWidget, title: str, key: str) -> None:
    '''Keep card title/marker in sync with row metadata from service rows.'''
    card.title_label.setText(title)
    card.set_marker_text(f"[{key}]")


def _attach_widget_to_card(
    widget: OptoControlManagerWidget,
    state_obj: BaseOptoControl,
    card: InstanceCardWidget,
) -> bool:
    '''
    Resolve and mount a concrete control widget for an expanded card.

    Route:
    - UI expand event -> this method -> OptoControlService.get_widget
    - concrete control widget returned and inserted into existing card body layout.
    '''
    try:
        child_widget = widget.opto_control_service.get_widget(
            state_obj,
            parent=card.body_container,
            display_service=widget.display_service,
            on_change=lambda c=state_obj, w=widget: _on_widget_changed(w, c),
        )
        card.set_body_widget(child_widget)
        return True
    except Exception as exc:
        card.setToolTip(f"Error: {exc}")
        return False


def _is_stale_widget(card: InstanceCardWidget) -> bool:
    '''Probe cached QWidget before trusting it after incremental updates.'''
    if card.body_layout.count() == 0:
        return False
    widget_item = card.body_layout.itemAt(0)
    if widget_item is None:
        return False
    body_widget = widget_item.widget()
    if body_widget is None:
        return False
    try:
        body_widget.isVisible()
    except Exception as exc:
        card.setToolTip(f"Warning: stale body widget detected ({exc})")
        return True
    return False


def _reset_control_widget_cache(state_obj: BaseOptoControl) -> None:
    '''
    Clear any cached widget reference on a control instance.

    This lets a stale deleted QWidget wrapper be rebuilt once when retrying attach.
    '''
    if hasattr(state_obj, "widget"):
        try:
            setattr(state_obj, "widget", None)
        except Exception:
            pass


def _on_widget_changed(widget: OptoControlManagerWidget, state_obj: BaseOptoControl) -> None:
    widget.opto_control_service.mark_control_changed(state_obj)


def _reorder_cards(
    widget: OptoControlManagerWidget,
    desired_cards: dict[BaseOptoControl, InstanceCardWidget],
    rows: list[dict[str, object]],
) -> None:
    '''
    Preserve service ordering while reusing existing card objects.

    We intentionally detach cards from layout and insert again to change order
    without destroying card instances that may hold expanded child widgets.
    '''
    desired_order = [row["state"] for row in rows]

    while widget.instances_layout.count() > 0:
        item = widget.instances_layout.takeAt(0)
        card = item.widget() #pyright: ignore
        if card is not None:
            card.setParent(None)

    for state in desired_order:
        card = desired_cards.get(state)
        if card is not None:
            widget.instances_layout.insertWidget(widget.instances_layout.count(), card)



from __future__ import annotations

from dataclasses import dataclass

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QComboBox,
    QFrame,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QScrollArea,
    QVBoxLayout,
    QWidget,
)


@dataclass
class InstrumentManagerUI:
    type_combo: QComboBox
    add_btn: QPushButton
    status_label: QLabel
    instances_scroll: QScrollArea
    instances_content: QWidget
    instances_layout: QVBoxLayout


def build_instrument_manager_ui(owner: QWidget) -> InstrumentManagerUI:
    root = QVBoxLayout(owner)
    root.setContentsMargins(0, 0, 0, 0)
    root.setSpacing(8)

    add_row = QHBoxLayout()
    add_row.setContentsMargins(0, 0, 0, 0)
    add_row.addWidget(QLabel("Instrument:", owner))
    type_combo = QComboBox(owner)
    add_row.addWidget(type_combo, 1)
    add_btn = QPushButton("Add", owner)
    add_row.addWidget(add_btn)
    root.addLayout(add_row)

    status_label = QLabel("Status: ready", owner)
    root.addWidget(status_label)

    instances_scroll = QScrollArea(owner)
    instances_scroll.setWidgetResizable(True)
    instances_scroll.setFrameShape(QFrame.Shape.NoFrame)
    instances_scroll.setObjectName("instrumentScroll")
    instances_scroll.setStyleSheet(
        "#instrumentScroll {"
        "background: palette(alternate-base);"
        "border: 1px solid palette(mid);"
        "border-radius: 8px;"
        "}"
        "#instrumentScroll > QWidget > QWidget {"
        "background: palette(alternate-base);"
        "}"
    )
    instances_content = QWidget(instances_scroll)
    instances_layout = QVBoxLayout(instances_content)
    instances_layout.setContentsMargins(8, 8, 8, 8)
    instances_layout.setSpacing(4)
    instances_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
    instances_scroll.setWidget(instances_content)
    root.addWidget(instances_scroll, 1)

    return InstrumentManagerUI(
        type_combo=type_combo,
        add_btn=add_btn,
        status_label=status_label,
        instances_scroll=instances_scroll,
        instances_content=instances_content,
        instances_layout=instances_layout,
    )



"""A measurement can be stopped, resumed, started for a defined period of time, or cleared."""

from time import sleep

from Swabian import TimeTagger

# Create a TimeTagger instance to control your hardware
tagger = TimeTagger.createTimeTagger()

# Activate test signal on channel 1
tagger.setTestSignal(1, True)

print("""
                        | getCountsTotal() | getCaptureDuration()
------------------------+------------------+---------------------""")

# We create a Countrate instance, it acquires data immediately
countrate = TimeTagger.Countrate(tagger=tagger, channels=[1])
sleep(.5)
print("(1) Init, sleep(0.5s)   | {:>16,d} | {:>20,d}".format(
    countrate.getCountsTotal()[0], countrate.getCaptureDuration()))

# Stop the measurement
countrate.stop()
sleep(.5)
print("(2) ... stop            | {:>16,d} | {:>20,d}".format(
    countrate.getCountsTotal()[0], countrate.getCaptureDuration()))

# By calling start, we can resume the measurement
countrate.start()
print("(3) ... resume          | {:>16,d} | {:>20,d}".format(
    countrate.getCountsTotal()[0], countrate.getCaptureDuration()))
sleep(.5)
print("(4) ... sleep(0.5s)     | {:>16,d} | {:>20,d}".format(
    countrate.getCountsTotal()[0], countrate.getCaptureDuration()))

# We can clear the acquired data while the measurement is running
countrate.clear()
sleep(.5)
print("(5) clear, sleep(0.5s)  | {:>16,d} | {:>20,d}".format(
    countrate.getCountsTotal()[0], countrate.getCaptureDuration()))
countrate.startFor(500000000000)
countrate.waitUntilFinished()

print("(6) startFor(0.5s)      | {:>16,d} | {:>20,d}".format(
    countrate.getCountsTotal()[0], countrate.getCaptureDuration()))

print("""
Please note the following:

(1),(2),(3)
The measurements can have the very same 'CaptureDuration' and Counts, but not necessarily.
The total acquisition time can be less than 0.5 s because the acquisition
does not start right away, but the initialization is non-blocking.

(5) vs (1)
The acquisition time can be the same, but not necessarily.

(6)
Only startFor() guarantees a defined acquisition time down to the ps.

How multiple measurements can be synchronized is shown in the following example
C_synchronized_measurements_histogramming.""")

TimeTagger.freeTimeTagger(tagger)



"""In this example, we look at methods that provide information on the Time Tagger(s) in your lab."""

import pprint

from Swabian import TimeTagger

available_taggers = TimeTagger.scanTimeTagger()
if available_taggers:
    print("\nTime Taggers available via TimeTagger.scanTimeTagger():")
    print(available_taggers)
else:
    print("There are no Time Taggers available. Connect one and retry.")
    exit()

tagger = TimeTagger.createTimeTagger()

print("\nYour Time Tagger model via tagger.getModel():")
print(tagger.getModel())

print("\nYour Time Tagger serial number via tagger.getSerial():")
print(tagger.getSerial())

print("\nAcquiring internal sensor data of your Time Tagger via tagger.getSensorData().")
try:
    print("\nSensor data:")
    pprint.pprint(tagger.getSensorData())
except RuntimeError:
    print("\nThe {}s do not support sensor readout.".format(tagger.getModel()))

print("\nFull Time Tagger Configuration via tagger.getConfiguration():")
pprint.pprint(tagger.getConfiguration())

TimeTagger.freeTimeTagger(tagger)



from __future__ import annotations

from typing import TYPE_CHECKING

from PyQt6.QtWidgets import QMessageBox

from pyrpoc.gui.main_widgets.acquisition_mgr.forms import apply_values, build_param_form, collect_values

if TYPE_CHECKING:
    from pyrpoc.gui.main_widgets.acquisition_mgr.widget import AcquisitionManagerWidget


def populate_modalities(widget: AcquisitionManagerWidget) -> None:
    current = widget.modality_combo.currentText()
    widget.modality_combo.blockSignals(True)
    widget.modality_combo.clear()
    rows = widget.modality_service.list_available()
    for row in rows:
        widget.modality_combo.addItem(row["key"])
    widget.modality_combo.blockSignals(False)

    if current and widget.modality_combo.findText(current) >= 0:
        widget.modality_combo.setCurrentText(current)
    elif widget.modality_combo.count() > 0:
        widget.modality_combo.setCurrentIndex(0)
        on_modality_selected(widget, widget.modality_combo.currentText())


def on_modality_selected(widget: AcquisitionManagerWidget, key: str) -> None:
    if not key:
        return
    try:
        widget.modality_service.select_modality(key)
    except Exception as exc:
        handle_error(widget, str(exc))


def handle_modality_selected(widget: AcquisitionManagerWidget, key: str) -> None:
    if key:
        index = widget.modality_combo.findText(key)
        if index >= 0 and widget.modality_combo.currentIndex() != index:
            widget.modality_combo.blockSignals(True)
            widget.modality_combo.setCurrentIndex(index)
            widget.modality_combo.blockSignals(False)

    parameter_groups = widget.modality_service.get_selected_parameters()
    existing_values = widget.modality_service.get_parameter_values()
    build_param_form(
        widget.ui,
        widget.state,
        parameter_groups,
        initial_values=existing_values,
        on_change=lambda w=widget: on_parameter_widgets_changed(w),
    )
    if existing_values:
        on_parameter_widgets_changed(widget)


def configure_modality(widget: AcquisitionManagerWidget) -> None:
    if widget.modality_service.app_state.modality.selected_class is None:
        key = widget.modality_combo.currentText().strip()
        if key:
            widget.modality_service.select_modality(key)
    params = collect_values(widget.state.param_widgets)
    widget.modality_service.set_parameter_values(params)
    widget.modality_service.configure(params)


def on_start_clicked(widget: AcquisitionManagerWidget) -> None:
    try:
        configure_modality(widget)
        widget.modality_service.start()
    except Exception as exc:
        handle_error(widget, str(exc))


def on_continuous_clicked(widget: AcquisitionManagerWidget) -> None:
    try:
        configure_modality(widget)
        widget.modality_service.start(continuous=True)
        widget.status_label.setText("Status: continuous acquisition")
    except Exception as exc:
        handle_error(widget, str(exc))


def on_stop_clicked(widget: AcquisitionManagerWidget) -> None:
    try:
        widget.modality_service.stop()
    except Exception as exc:
        handle_error(widget, str(exc))


def handle_requirements_changed(
    widget: AcquisitionManagerWidget,
    ok: bool,
    missing_names: list[str],
) -> None:
    if ok:
        widget.status_label.setText("Status: ready")
        return
    text = ", ".join(missing_names)
    widget.status_label.setText(f"Status: missing instruments -> {text}")


def handle_error(widget: AcquisitionManagerWidget, message: str) -> None:
    widget.status_label.setText(f"Status: error - {message}")
    QMessageBox.critical(widget, "Acquisition Error", message)


def on_service_error(widget: AcquisitionManagerWidget, message: str) -> None:
    widget.status_label.setText(f"Status: error - {message}")


def on_parameter_widgets_changed(widget: AcquisitionManagerWidget) -> None:
    values = collect_values(widget.state.param_widgets)
    widget.modality_service.set_parameter_values(values)


def on_parameter_values_changed(widget: AcquisitionManagerWidget, values: object) -> None:
    if not isinstance(values, dict):
        return
    apply_values(widget.state.param_widgets, values)



from .segmentation_methods import labels_to_contours, segment
from .editor import RPOCMaskEditor
from .types import RPOCEditorState, RPOCImageInput, RPOCRoi

__all__ = [
    "segment",
    "labels_to_contours",
    "RPOCMaskEditor",
    "RPOCImageInput",
    "RPOCRoi",
    "RPOCEditorState",
]



from __future__ import annotations

from dataclasses import dataclass

from PyQt6.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QPushButton,
    QVBoxLayout,
    QWidget,
)


@dataclass
class DisplayManagerUI:
    display_combo: QComboBox
    name_input: QLineEdit
    add_btn: QPushButton
    instances_list: QListWidget
    attach_btn: QPushButton
    detach_btn: QPushButton
    remove_btn: QPushButton
    status_label: QLabel


def build_display_manager_ui(owner: QWidget) -> DisplayManagerUI:
    root = QVBoxLayout(owner)

    add_row = QHBoxLayout()
    add_row.addWidget(QLabel("Display:", owner))
    display_combo = QComboBox(owner)
    add_row.addWidget(display_combo, 1)
    add_btn = QPushButton("Add", owner)
    add_row.addWidget(add_btn, 0)
    root.addLayout(add_row)

    name_row = QHBoxLayout()
    name_row.addWidget(QLabel("Display Name:", owner))
    name_input = QLineEdit(owner)
    name_input.setPlaceholderText("Optional display name")
    name_row.addWidget(name_input, 1)
    root.addLayout(name_row)

    root.addWidget(QLabel("Active Displays:", owner))
    instances_list = QListWidget(owner)
    root.addWidget(instances_list, 1)

    action_row = QHBoxLayout()
    attach_btn = QPushButton("Attach", owner)
    detach_btn = QPushButton("Detach", owner)
    remove_btn = QPushButton("Remove", owner)
    action_row.addWidget(attach_btn)
    action_row.addWidget(detach_btn)
    action_row.addWidget(remove_btn)
    action_row.addStretch(1)
    root.addLayout(action_row)

    status_label = QLabel("Status: ready", owner)
    root.addWidget(status_label)

    # Remove per-widget display tabs; display widgets now live in main ADS docks.
    return DisplayManagerUI(
        display_combo=display_combo,
        name_input=name_input,
        add_btn=add_btn,
        instances_list=instances_list,
        attach_btn=attach_btn,
        detach_btn=detach_btn,
        remove_btn=remove_btn,
        status_label=status_label,
    )



from __future__ import annotations

from typing import Any

import numpy as np
from PyQt6.QtCore import QObject, pyqtSignal

from pyrpoc.backend_utils.array_contracts import infer_array_contract
from pyrpoc.backend_utils.parameter_utils import ParameterValidationError, coerce_parameter_values
from pyrpoc.domain.app_state import AppState, ParameterValue
from pyrpoc.displays.base_display import BaseDisplay
from pyrpoc.displays.display_registry import display_registry
from pyrpoc.rpoc.types import RPOCImageInput


class DisplayService(QObject):
    # from display manager actions / modality data stream
    # through service inventory + push_data routing
    # to live display widgets and UI refresh/autosave signals.
    display_added = pyqtSignal(object)
    display_removed = pyqtSignal(object)
    display_changed = pyqtSignal(object)
    display_error = pyqtSignal(object, str)

    def __init__(self, app_state: AppState, parent=None):
        super().__init__(parent)
        self.app_state = app_state
        self._reported_incompatibilities: set[tuple[int, str]] = set()
        self._last_data: np.ndarray | None = None

    def list_available(self) -> list[dict[str, Any]]:
        return display_registry.describe_all()

    def list_compatible_with(self, data_contract: str) -> list[str]:
        compatible: list[str] = []
        if not isinstance(data_contract, str) or not data_contract.strip():
            return compatible
        for key in display_registry.list_keys():
            display_cls = display_registry.get_class(key)
            accepted_contracts = getattr(display_cls, "ACCEPTED_DATA_CONTRACTS", [])
            if data_contract in accepted_contracts:
                compatible.append(key)
        return compatible

    def create_display(
        self,
        key: str,
        raw_settings: dict[str, Any] | None = None,
        user_label: str | None = None,
        *,
        instance_id: str | None = None,
        persisted_state: dict[str, Any] | None = None,
        attached: bool = True,
        dock_visible: bool = True,
    ) -> BaseDisplay:
        """
        Build one display instance, configure it, and register it in runtime inventory.

        Route:
        - DisplayManager add click
        - -> `display_mgr.handlers.on_add_clicked`
        - -> this method
        - -> `display_added` signal for tab/list rendering.
        """
        display_cls = display_registry.get_class(key)
        settings_parameters = display_cls.DISPLAY_PARAMETERS
        settings_input = raw_settings or {}

        try:
            settings = coerce_parameter_values(settings_parameters, settings_input)
            widget = display_cls()
            widget.configure(settings)
        except (ParameterValidationError, Exception) as exc:
            self.display_error.emit(None, str(exc))
            raise

        widget.attached = bool(attached)
        widget.docked_visible = bool(dock_visible)
        if instance_id:
            widget.instance_id = str(instance_id)
        widget.config_values = [ParameterValue(label=k, value=v) for k, v in settings.items()]
        widget.last_error = None
        widget.user_label = user_label if isinstance(user_label, str) and user_label.strip() else getattr(
            widget,
            "user_label",
            None,
        )
        if isinstance(persisted_state, dict):
            widget.import_persistence_state(dict(persisted_state))

        self.app_state.displays.append(widget)
        widget.set_persist_callback(lambda display=widget: self.mark_display_changed(display))
        self.display_added.emit(widget)
        return widget

    def remove_display(self, display: BaseDisplay) -> None:
        if display not in self.app_state.displays:
            return
        self.app_state.displays.remove(display)
        display.set_persist_callback(None)
        self.display_removed.emit(display)
        # `display_removed` listeners (MainGUI) own dock detachment.
        # Defer widget deletion until after those slots finish.
        display.deleteLater()

    def attach(self, display: BaseDisplay) -> None:
        self._require_state(display)
        display.attached = True
        self.display_changed.emit(display)

    def detach(self, display: BaseDisplay) -> None:
        self._require_state(display)
        display.attached = False
        self.display_changed.emit(display)

    def set_dock_visibility(self, display: BaseDisplay, visible: bool) -> None:
        if display not in self.app_state.displays:
            return
        self._require_state(display)
        if bool(getattr(display, "docked_visible", True)) == bool(visible):
            return
        display.docked_visible = bool(visible)
        self.display_changed.emit(display)

    def push_data(self, data: np.ndarray) -> None:
        """
        Fan out one modality frame to all attached compatible displays.

        Route:
        - `AppController` wires `ModalityService.data_ready`
        - -> this method
        - -> per-display render calls + error signals.
        """
        if not isinstance(data, np.ndarray):
            self.display_error.emit(None, f"display stream expects numpy.ndarray, got {type(data).__name__}")
            return

        self._last_data = data
        payload_contract = infer_array_contract(data)
        for display in self.app_state.displays:
            if not getattr(display, "attached", True):
                continue
            if not bool(getattr(display, "docked_visible", True)):
                continue

            accepted_contracts = getattr(display, "ACCEPTED_DATA_CONTRACTS", [])
            compatible = payload_contract is not None and payload_contract in accepted_contracts
            if not compatible:
                key = (id(display), str(payload_contract))
                if key not in self._reported_incompatibilities:
                    self._reported_incompatibilities.add(key)
                    self.display_error.emit(
                        display,
                        (
                            f"display cannot render payload contract "
                            f"{payload_contract!r}; accepted={accepted_contracts}"
                        ),
                    )
                continue

            try:
                display.render(data)
            except Exception as exc:
                display.last_error = str(exc)
                self.display_error.emit(display, str(exc))

    def get_latest_data(self) -> np.ndarray | None:
        return self._last_data

    def list_instances(self) -> list[dict[str, Any]]:
        rows: list[dict[str, Any]] = []
        for display in self.app_state.displays:
            key = display.type_key
            cls = display_registry.get_class(key)
            name = getattr(display, "user_label", None) or getattr(cls, "DISPLAY_NAME", key)
            rows.append(
                {
                    "state": display,
                    "display_id": id(display),
                    "key": key,
                    "name": name,
                    "attached": bool(getattr(display, "attached", True)),
                    "docked_visible": bool(getattr(display, "docked_visible", True)),
                }
            )
        return rows

    def get_display_by_id(self, display_id: int) -> BaseDisplay | None:
        for display in self.app_state.displays:
            if id(display) == display_id:
                return display
        return None

    def set_display_name(self, display: BaseDisplay, user_label: str) -> None:
        self._require_state(display)
        label = (user_label or "").strip()
        display.user_label = label or None
        self.display_changed.emit(display)

    def mark_display_changed(self, display: BaseDisplay) -> None:
        self._require_state(display)
        self.display_changed.emit(display)

    def get_widget(self, display: BaseDisplay) -> BaseDisplay:
        self._require_state(display)
        return display

    def get_rpoc_input(self, display: BaseDisplay) -> RPOCImageInput | None:
        widget = self.get_widget(display)
        exporter = getattr(widget, "export_rpoc_input", None)
        if not callable(exporter):
            return None
        payload = exporter()
        if payload is None:
            return None
        if not isinstance(payload, RPOCImageInput):
            raise TypeError(
                f"display export_rpoc_input returned {type(payload).__name__}, "
                "expected RPOCImageInput or None"
            )
        return payload

    def _require_state(self, display: BaseDisplay) -> None:
        if display not in self.app_state.displays:
            raise KeyError("display does not exist")

    def clear_all(self) -> None:
        # from session reset/restore -> through clear_all -> to empty display inventory
        for display in list(self.app_state.displays):
            self.remove_display(display)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
import threading
from typing import Any

import numpy as np

from pyrpoc.backend_utils.array_contracts import CONTRACT_CHW_FLOAT32
from pyrpoc.backend_utils.contracts import ParameterGroups
from pyrpoc.backend_utils.parameter_utils import validate_parameter_groups
from pyrpoc.instruments.base_instrument import BaseInstrument
from pyrpoc.optocontrols.base_optocontrol import BaseOptoControl


class BaseModality(ABC):
    MODALITY_KEY: str = "base_modality"
    DISPLAY_NAME: str = "Base Modality"
    PARAMETERS: ParameterGroups = {}
    REQUIRED_INSTRUMENTS: list[type[BaseInstrument]] = []
    OPTIONAL_INSTRUMENTS: list[type[BaseInstrument]] = []
    ALLOWED_OPTOCONTROLS: list[type[BaseOptoControl]] = []
    OUTPUT_DATA_CONTRACT: str = CONTRACT_CHW_FLOAT32
    ALLOWED_DISPLAYS: list[str] = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        required = getattr(cls, "REQUIRED_INSTRUMENTS", [])
        optional = getattr(cls, "OPTIONAL_INSTRUMENTS", [])
        allowed_opto = getattr(cls, "ALLOWED_OPTOCONTROLS", [])

        if not isinstance(required, list):
            raise TypeError("REQUIRED_INSTRUMENTS must be a list")
        if not isinstance(optional, list):
            raise TypeError("OPTIONAL_INSTRUMENTS must be a list")
        if not isinstance(allowed_opto, list):
            raise TypeError("ALLOWED_OPTOCONTROLS must be a list")

        for instrument_cls in [*required, *optional]:
            if not isinstance(instrument_cls, type) or not issubclass(instrument_cls, BaseInstrument):
                raise TypeError(
                    f"{cls.__name__} instrument requirements must contain BaseInstrument subclasses"
                )
        for optocontrol_cls in allowed_opto:
            if not isinstance(optocontrol_cls, type) or not issubclass(optocontrol_cls, BaseOptoControl):
                raise TypeError(
                    f"{cls.__name__} ALLOWED_OPTOCONTROLS must contain BaseOptoControl subclasses"
                )

        validate_parameter_groups(getattr(cls, "PARAMETERS", {}))

        output_contract = getattr(cls, "OUTPUT_DATA_CONTRACT", CONTRACT_CHW_FLOAT32)
        if not isinstance(output_contract, str) or not output_contract.strip():
            raise TypeError("OUTPUT_DATA_CONTRACT must be a non-empty string")

    def __init__(self):
        self._running = False
        self._configured = False
        self._params: dict[str, Any] = {}
        self._instruments: dict[type[BaseInstrument], BaseInstrument] = {}
        self._opto_controls: list[BaseOptoControl] = []

    @classmethod
    def get_contract(cls) -> dict[str, Any]:
        return {
            "modality_key": cls.MODALITY_KEY,
            "display_name": cls.DISPLAY_NAME,
            "parameters": cls.PARAMETERS,
            "required_instruments": cls.REQUIRED_INSTRUMENTS,
            "optional_instruments": cls.OPTIONAL_INSTRUMENTS,
            "allowed_optocontrols": cls.ALLOWED_OPTOCONTROLS,
            "output_data_contract": cls.OUTPUT_DATA_CONTRACT,
            "allowed_displays": cls.ALLOWED_DISPLAYS,
        }

    @abstractmethod
    def configure(
        self,
        params: dict[str, Any],
        instruments: dict[type[BaseInstrument], BaseInstrument],
        opto_controls: list[BaseOptoControl],
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def start(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def acquire_once(self) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def prepare_acquisition_storage(self, *, frame_limit: int | None) -> None:
        raise NotImplementedError

    @abstractmethod
    def save_acquired_frame(self, frame: np.ndarray, *, frame_index: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def finalize_acquisition_storage(
        self,
        *,
        frame_count: int,
        frame_limit: int | None,
        error: Exception | None,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_frame_limit(self) -> int:
        raise NotImplementedError

    def get_active_channel_labels(self) -> list[str]:
        return []

    def run_acquisition_threaded(
        self,
        on_frame: Callable[[np.ndarray], None],
        *,
        frame_limit: int | None = None,
        should_stop: Callable[[], bool],
        on_error: Callable[[Exception], None],
        on_finished: Callable[[int, Exception | None], None],
    ) -> threading.Thread:

        def worker() -> None:
            acquired = 0
            error_seen = None
            try:
                self.start()
                while not (should_stop() if should_stop else False):
                    frame = self.acquire_once()
                    acquired += 1
                    on_frame(frame)
                    if frame_limit is not None and acquired >= frame_limit:
                        break
            except Exception as exc: 
                error_seen = exc
                on_error(exc)
            finally:
                try:
                    self.stop()
                except Exception as stop_exc:
                    if error_seen is None:
                        error_seen = stop_exc
                        on_error(stop_exc)
                if on_finished is not None:
                    on_finished(acquired, error_seen)

        thread = threading.Thread(target=worker, daemon=True)
        thread.start()
        return thread



import numba
import numpy

from Swabian import TimeTagger


# Create a new IntensityRenderer based on the TimeDifferencesRender backend class
class IntensityRenderer(TimeTagger.FlimBase):

    def __init__(self, tagger, start_channel, click_channel, pixel_begin_channel, pixels_x, pixels_y, n_bins, binwidth, pixel_end_channel=TimeTagger.CHANNEL_UNUSED, frame_begin_channel=TimeTagger.CHANNEL_UNUSED, num_frames=0, divider=1, is_resonant=False):
        TimeTagger.FlimBase.__init__(self, tagger, start_channel, click_channel, pixel_begin_channel,
                                     pixels_x * pixels_y, n_bins, binwidth, pixel_end_channel, frame_begin_channel, num_frames, divider, False)
        self.pixels_x = pixels_x
        self.pixels_y = pixels_y
        self.frames = []
        self.bins = n_bins
        self.is_resonant = is_resonant

        # When inheriting from a measurement, initialization must be done manually and after all
        # needed variables/resources are initialized.
        self.initialize()

    # render will be called by the backend every time it finishes a frame, the backend
    # will pass the current histogram as read-only(shouldn't be modified). This method
    # will then use the current histogram to generate a frame and store it. Due to the speed
    # needs of this class, it is suggested to break it into smaller functions and use numba, so
    # most hard processing is done as fast as possible
    def frameReady(self, frame_number, frame_data, pixels_start_times, pixels_end_times, frame_start_time, frame_end_time):
        # the received data isd a 1 dimensional array of [pixels_x * pixels_y * bins]
        # we must reshape it.
        alt_data = frame_data.reshape(self.pixels_y, self.pixels_x, self.bins)
        alt_pixel_start_times = pixels_start_times.reshape(
            self.pixels_y, self.pixels_x)
        alt_pixel_end_times = pixels_end_times.reshape(self.pixels_y, self.pixels_x)
        frame = numpy.empty([self.pixels_y - 1, self.pixels_x - 1])
        # intensity is calculated differently in resonant and non resonant scan,
        # for resonant we must alternate the lines direction every odd line.
        if (not self.is_resonant):
            IntensityRenderer.__calculate_intensity(
                alt_data, frame, alt_pixel_start_times, alt_pixel_end_times, False)
        else:
            IntensityRenderer.__calculate_intensity(
                alt_data, frame, alt_pixel_start_times, alt_pixel_end_times, True)
        self.frames.append(frame)

    def getFrames(self):
        return self.frames

    def hasFrames(self):
        return len(self.frames) != 0

    def getLastFrame(self):
        return self.frames[-1]

    # This function calculates the intensity as the sum of all the histogram.
    # Special care has been taken in order to make sure it compiles in numba.
    # Among those things: set it as static method, only use numpy arrays,
    # don't use lists, dictionaries or complex python types, etc.
    @staticmethod
    @numba.jit(nopython=True, nogil=True)
    def __calculate_intensity(data, frame, start_pixel_times, end_pixel_times, flip):
        for y in range(1, data.shape[0]):
            if ((not flip) or y % 2 == 0):
                for x in range(1, data.shape[1]):
                    sum = 0
                    for b in range(0, data.shape[2]):
                        sum = sum + data[y][x][b]
                    frame[y - 1][x - 1] = (end_pixel_times[y]
                                           [x] - start_pixel_times[y][x]) * sum
            else:
                for x in range(data.shape[1] - 1, 0, -1):
                    sum = 0
                    for b in range(0, data.shape[2]):
                        sum = sum + data[y][x][b]
                    frame[y - 1][frame.shape[1] - x] = (end_pixel_times[y][x] - start_pixel_times[y][x]) * sum



from __future__ import annotations

from collections.abc import Callable

from PyQt6.QtWidgets import QWidget

from pyrpoc.instruments.base_instrument import BaseInstrument, BaseInstrumentWidget
from pyrpoc.instruments.instrument_registry import instrument_registry
from pyrpoc.instruments.instrument_widgets import ConfocalDAQInstrumentWidget


@instrument_registry.register("confocal_daq")
class ConfocalDAQInstrument(BaseInstrument):
    INSTRUMENT_KEY = "confocal_daq"
    DISPLAY_NAME = "Confocal DAQ"

    def __init__(
        self,
        alias: str | None = None,
        *,
        instance_id: str | None = None,
        user_label: str | None = None,
        connected: bool = False,
    ):
        super().__init__(
            alias=alias,
            instance_id=instance_id,
            user_label=user_label,
            connected=connected,
        )
        self.device_name: str = "Dev1"
        self.sample_rate_hz: float = 100_000.0
        self.ai_channel_numbers: list[int] = list(range(9))
        self.active_ai_channels: list[bool] = [True] * 9
        self.fast_axis_ao: int = 0
        self.slow_axis_ao: int = 1

        self.widget: BaseInstrumentWidget | None = None

    def get_widget(
        self,
        parent: QWidget | None = None,
        on_change: Callable[[], None] | None = None,
    ) -> BaseInstrumentWidget:
        if self.widget is None:
            self.widget = ConfocalDAQInstrumentWidget(self, on_change=on_change, parent=parent)
        elif parent is not None:
            self.widget.setParent(parent)
        if on_change is not None:
            self.widget.on_change = on_change
        return self.widget

    def prepare_for_acquisition(self) -> tuple[object, ...]:
        active_ai = [
            ai_idx
            for ai_idx, enabled in zip(self.ai_channel_numbers, self.active_ai_channels, strict=False)
            if enabled
        ]
        return (
            self.device_name,
            float(self.sample_rate_hz),
            len(self.ai_channel_numbers),
            list(self.ai_channel_numbers),
            active_ai,
            int(self.fast_axis_ao),
            int(self.slow_axis_ao),
        )

    def get_collapsed_summary(self) -> str:
        active_ai = [
            ai_idx
            for ai_idx, enabled in zip(self.ai_channel_numbers, self.active_ai_channels, strict=False)
            if enabled
        ]
        ai_text = ", ".join(str(ai) for ai in active_ai) if active_ai else "none"
        return f"AIs: {ai_text}; Galvos: {int(self.fast_axis_ao)},{int(self.slow_axis_ao)}"


ConfocalDAQ = ConfocalDAQInstrument



"""This script serves to calibrate the measurements for counting the number of
photons.
The results can be used with the TimeTagger.Experimental.PhotonNumber virtual channel.

Here, the Histogram2D measurement will take the information on the time
differences between the trigger signal and the front and rear edges of the
resulting signal.
The typical distribution of the points is tilted to obtain the best condition
to observe the separation between the peaks associated with the photon numbers.

The results necessary for the virtual channel are:
x_intercepts: corresponds to the series of points separating the regions associated
with the different number of photons. The series is defined as the intercept
on the axis of the front edges of the lines separating the previously defined
regions in the original 2D Histogram
slope: corresponds to the slope of the tilt of the 2D histogram.
"""

from json import dump
import matplotlib.pyplot as plt
import numpy as np
from pnr_constants import FRONT_EDGE, REAR_EDGE, LASER_CHANNEL, FOLDER, FILENAME
from scipy.signal import find_peaks

from Swabian import TimeTagger


def tilt(original_data, slope):
    # Rotate the image, performing a shift line by line of the 2D histogram
    tilted_data = np.zeros((n_bins_rear, n_bins_front), dtype=float)
    for i in range(1, n_bins_rear):
        shift = slope * i
        index_shift = int(np.floor(shift))
        fractional_shift = shift - index_shift
        tilted_data[i, index_shift:-1] = fractional_shift*original_data[i,
                                                                        :(n_bins_front-index_shift-1)] + (1-fractional_shift)*original_data[i, 1:(n_bins_front-index_shift)]
    return tilted_data


# Input parameters
n_bins_front = 1000
w_bins_front = 1
# Preliminary numbers. They'll be optimized below
n_bins_rear = 5000
w_bins_rear = 1


# Create Time Tagger object
tagger = TimeTagger.createTimeTaggerVirtual(filename=FOLDER + FILENAME)

# Find the maximum position of front edges to shift the laser trigger to the
# front
start_stop_front = TimeTagger.StartStop(tagger=tagger,
                                        click_channel=LASER_CHANNEL,
                                        start_channel=FRONT_EDGE,
                                        binwidth=w_bins_front)
# Find the maximum position of rear edges to adjust binning
start_stop_rear = TimeTagger.StartStop(tagger=tagger,
                                       click_channel=LASER_CHANNEL,
                                       start_channel=REAR_EDGE,
                                       binwidth=w_bins_rear)

tagger.run()
tagger.waitUntilFinished()

# Front maximum position
front_data = start_stop_front.getData()
front_center = front_data[front_data[:, 1].argmax(), 0]
delay = front_center + (n_bins_front*w_bins_front)//2
tagger.setInputDelay(LASER_CHANNEL, -delay)
# Rear maximum position
rear_data = start_stop_rear.getData()
rear_center = rear_data[rear_data[:, 1].argmax(), 0]
# new position after the laser negative delay
rear_peak = delay - rear_center

# Adjust rear edge binning
# check if rear peak is inside the binning, otherwise adjust the width to avoid having a massive matrix
n_bin_max = 5000
iter = 0
if n_bins_rear < rear_peak:
    n_bins_tmp = rear_peak
    while n_bin_max < n_bins_tmp:
        iter += 1
        n_bins_tmp = round(rear_peak + 5000, -4) // (iter*2)
    n_bins_rear = n_bins_tmp
    w_bins_rear = iter*2

# Measurement of the 2d histogram
data = TimeTagger.Histogram2D(
    tagger, LASER_CHANNEL, REAR_EDGE, FRONT_EDGE, w_bins_rear, w_bins_front, n_bins_rear, n_bins_front)

tagger.appendFile(file=FOLDER + FILENAME)

# Data loading
while not tagger.waitUntilFinished(timeout=0):
    plt.figure(1)
    plt.pcolormesh(data.getIndex_2(), data.getIndex_1(),
                   data.getData(), shading="auto")

    plt.pause(.1)


plt.figure(1)
plt.colorbar()
plt.title('Hist 2D, before angle correction')
plt.xlabel('Trigger, CH'+str(REAR_EDGE) + ' front edge [ps]')
plt.ylabel('Trigger, CH'+str(abs(FRONT_EDGE)) + ' rear edge [ps]')
plt.set_cmap('jet')


# Rotation of the 2D histogram to find the best conditions to distinguish the
# peaks
raw_data = np.array(data.getData(), dtype=float)
sum_squared = np.sum((np.sum(raw_data, axis=0) ** 2))
best_slope = 0

for slope in np.linspace(1, 1000, 1000)/(10000/w_bins_rear):
    # Evaluation of the best conditions to emphasize the separation between
    # the peaks, and distinguish better the photon number contribution.
    sum_squared_new = np.sum((np.sum(tilt(raw_data, slope), axis=0) ** 2))
    if sum_squared_new <= sum_squared:
        break
    sum_squared = sum_squared_new
    best_slope = slope

# Find local minima
tilted_data = tilt(raw_data, best_slope)
front_opt = np.sum(tilted_data, axis=0)
maxima_front, prop = find_peaks(front_opt, height=(
    max((front_opt)) + min((front_opt)))/4, distance=10)
x_intercepts = list()
for k in range(np.size(maxima_front)-1, 0, -1):
    avg = (maxima_front[k] + maxima_front[k-1])/2
    x_intercepts.append((1 + best_slope / w_bins_rear) * delay - avg)
x_intercepts.insert(0, x_intercepts[0] - 200)
x_intercepts.append(x_intercepts[-1] + 200)

with open(FOLDER + "parameters_ch"+str(abs(FRONT_EDGE))+".json", "w") as param_file:
    dump(dict(x_intercepts=x_intercepts,
         slope=best_slope / w_bins_rear, delay=int(delay)), param_file)
    print(
        f'Parameters exported to {FOLDER}parameters_ch{abs(FRONT_EDGE)}.json')

# Plot of the comparison of histograms before and after the optimization. The
# histograms are obtained after the integration along the rear-edge's axis
plt.figure(2)
plt.plot(data.getIndex_2(), np.sum(raw_data, axis=0), label='Original')
plt.plot(data.getIndex_2(), front_opt, label='Optimized')
plt.legend()
plt.xlabel(f'Trigger, CH{REAR_EDGE} front edge [ps]')
plt.title(f'Hist,  click CH {LASER_CHANNEL},  stop CH {FRONT_EDGE}')


# Plot of the 2D Histogram after rotation
plt.figure(3)

plt.pcolormesh(data.getIndex_2(), data.getIndex_1(),
               tilted_data, shading="auto")
plt.title('Hist 2D, after angle correction')
plt.xlabel(f'Trigger, CH{REAR_EDGE} front edge [ps]')
plt.ylabel(f'Trigger, CH{abs(FRONT_EDGE)} rear edge [ps]')
plt.colorbar()

plt.show()

TimeTagger.freeTimeTagger(tagger)



from matplotlib import pyplot as plt
import numba
import numpy as np

from Swabian import TimeTagger


class CustomStartMultipleStop(TimeTagger.CustomMeasurement):
    """
    Example for a single start - multiple stop measurement.
        The class shows how to access the raw time-tag stream.
    """

    def __init__(self, tagger, click_channel, start_channel, binwidth, n_bins):
        TimeTagger.CustomMeasurement.__init__(self, tagger)
        self.click_channel = click_channel
        self.start_channel = start_channel
        self.binwidth = binwidth
        self.max_bins = n_bins

        # The method register_channel(channel) activates
        # data transmission from the Time Tagger to the PC 
        # for the respective channels.
        self.register_channel(channel=click_channel)
        self.register_channel(channel=start_channel)

        self.clear_impl()

        # At the end of a CustomMeasurement construction,
        # we must indicate that we have finished.
        self.finalize_init()

    def __del__(self):
        # The measurement must be stopped before deconstruction to avoid
        # concurrent process() calls.
        self.stop()

    def getData(self):
        # Locking this instance to guarantee that process() is not running in parallel
        # This ensures to return a consistent data.
        with self.mutex:
            return self.data.copy()

    def getIndex(self):
        # This method does not depend on the internal state, so there is no
        # need for a lock.
        arr = np.arange(0, self.max_bins) * self.binwidth
        return arr

    def clear_impl(self):
        # The lock is already acquired within the backend.
        self.last_start_timestamp = 0
        self.data = np.zeros((self.max_bins,), dtype=np.uint64)

    def on_start(self):
        # The lock is already acquired within the backend.
        pass

    def on_stop(self):
        # The lock is already acquired within the backend.
        pass

    @staticmethod
    @numba.jit(nopython=True, nogil=True)
    def fast_process(
            tags,
            data,
            click_channel,
            start_channel,
            binwidth,
            last_start_timestamp):
        """
        A precompiled version of the histogram algorithm for better performance
        nopython=True: Only a subset of the python syntax is supported.
                       Avoid everything but primitives and numpy arrays.
                       All slow operation will yield an exception
        nogil=True:    This method will release the global interpreter lock. So
                       this method can run in parallel with other python code
        """
        for tag in tags:
            # tag.type can be: 0 - TimeTag, 1- Error, 2 - OverflowBegin, 3 -
            # OverflowEnd, 4 - MissedEvents (you can use the TimeTagger.TagType IntEnum)
            if tag['type'] != TimeTagger.TagType.TimeTag:
                # tag is not a TimeTag, so we are in an error state, e.g. overflow
                last_start_timestamp = 0
            elif tag['channel'] == click_channel and last_start_timestamp != 0:
                # valid event
                index = (tag['time'] - last_start_timestamp) // binwidth
                if index < data.shape[0]:
                    data[index] += 1
            if tag['channel'] == start_channel:
                last_start_timestamp = tag['time']
        return last_start_timestamp

    def process(self, incoming_tags, begin_time, end_time):
        """
        Main processing method for the incoming raw time-tags.

        The lock is already acquired within the backend.
        self.data is provided as reference, so it must not be accessed
        anywhere else without locking the mutex.

        Parameters
        ----------
        incoming_tags
            The incoming raw time tag stream provided as a read-only reference.
            The storage will be deallocated after this call, so you must not store a reference to
            this object. Make a copy instead.
            Please note that the time tag stream of all channels is passed to the process method,
            not only the ones from register_channel(...).
        begin_time
            Begin timestamp of the of the current data block.
        end_time
            End timestamp of the of the current data block.
        """
        self.last_start_timestamp = CustomStartMultipleStop.fast_process(
            incoming_tags,
            self.data,
            self.click_channel,
            self.start_channel,
            self.binwidth,
            self.last_start_timestamp)


# Channel definitions
CHAN_START = 1
CHAN_STOP = 2

if __name__ == '__main__':

    print("""Custom Measurement example

Implementation of a custom single start, multiple stop measurement, histogramming
the time differences of the two input channels.

The custom implementation will be compared to a the built-in Histogram class,
which is a multiple start, multiple stop measurement. But for the
selected time span of the histogram, multiple start does not make a difference.
""")

    tagger = TimeTagger.createTimeTagger()

    # enable the test signal
    tagger.setTestSignal([CHAN_START, CHAN_STOP], True)
    # delay the stop channel by 2 ns to make sure it is later than the start
    tagger.setInputDelay(CHAN_STOP, 2000)

    BINWIDTH = 1  # ps
    BINS = 4000

    # We first have to create a SynchronizedMeasurements object to synchronize several measurements
    with TimeTagger.SynchronizedMeasurements(tagger) as measurementGroup:
        # Instead of a real Time Tagger, we initialize the measurement with the proxy object measurementGroup.getTagger()
        # This adds the measurement to the measurementGroup. In contrast to a normal initialization of a measurement, the
        # measurement does not start immediately but waits for an explicit .start() or .startFor().
        custom_histogram = CustomStartMultipleStop(measurementGroup.getTagger(), CHAN_STOP, CHAN_START, binwidth=BINWIDTH, n_bins=BINS)
        histogram = TimeTagger.Histogram(measurementGroup.getTagger(), CHAN_STOP, CHAN_START, binwidth=BINWIDTH, n_bins=BINS)

        print("Acquire data...\n")
        measurementGroup.startFor(int(3e12))
        measurementGroup.waitUntilFinished()

        x_custom_histogram = custom_histogram.getIndex()
        y_custom_histogram = custom_histogram.getData()

        x_histogram = histogram.getIndex()
        y_histogram = histogram.getData()

        assert np.all(y_custom_histogram == y_histogram), "ERROR: The custom implementation has a different result than the Histogram from the API!"
        print("The custom implementation shows the very same data as the Histogram from the API.")

    TimeTagger.freeTimeTagger(tagger)

    plt.plot(x_custom_histogram, y_custom_histogram, label='custom histogram')
    plt.plot(x_histogram, y_histogram, '--', label='histogram')
    plt.xlabel('Time difference (ps)')
    plt.ylabel('Counts')
    plt.legend()
    plt.show()



from pyrpoc.backend_utils.registry import Registry
from .base_display import BaseDisplay


class DisplayRegistry(Registry):
    def __init__(self):
        super().__init__(name="DisplayRegistry", base_class=BaseDisplay)


display_registry = DisplayRegistry()



from __future__ import annotations

import threading
from typing import Any

import numpy as np
from PyQt6.QtCore import QObject, pyqtSignal

from pyrpoc.backend_utils.array_contracts import matches_array_contract
from pyrpoc.backend_utils.parameter_utils import coerce_parameter_values
from pyrpoc.domain.app_state import AppState, ParameterValue
from pyrpoc.instruments.base_instrument import BaseInstrument
from pyrpoc.modalities.base_modality import BaseModality
from pyrpoc.modalities.mod_registry import modality_registry
from pyrpoc.optocontrols.base_optocontrol import BaseOptoControl
from .instrument_service import InstrumentService


class ModalityService(QObject):
    modality_selected = pyqtSignal(str)
    modality_params_changed = pyqtSignal(object)
    requirements_changed = pyqtSignal(bool, list)
    acq_started = pyqtSignal()
    data_ready = pyqtSignal(object)
    acq_stopped = pyqtSignal()
    acq_error = pyqtSignal(str)

    def __init__(self, instrument_service: InstrumentService, app_state: AppState, parent=None):
        super().__init__(parent)
        self.instrument_service = instrument_service
        self.app_state = app_state
        self._active_frame_limit: int | None = 1
        self._active_output_contract: str = ""
        self._active_modality_instance: BaseModality | None = None
        self._frames_emitted: int = 0
        self.acquisition_thread: threading.Thread | None = None
        self.acquisition_stop_requested = threading.Event()
        self.acquisition_lock = threading.Lock()

    def list_available(self) -> list[dict[str, Any]]:
        return modality_registry.describe_all()

    def select_modality(self, key: str) -> None:
        if self.app_state.modality.running:
            self.stop()

        try:
            cls = modality_registry.get_class(key)
        except KeyError:
            self.app_state.modality.selected_key = None
            self.app_state.modality.selected_class = None
            self.app_state.modality.instance = None
            self.app_state.modality.configured_params = []
            self.acq_error.emit(f"unknown modality '{key}'")
            raise

        self.app_state.modality.selected_key = key
        self.app_state.modality.selected_class = cls
        self.app_state.modality.instance = cls()
        self.app_state.modality.configured_params = []
        self.modality_selected.emit(key)
        self.validate_required_instruments()

    def get_selected_parameters(self) -> dict[str, list]:
        if self.app_state.modality.selected_class is None:
            return {}
        return self.app_state.modality.selected_class.PARAMETERS

    def get_selected_contract(self) -> dict[str, Any]:
        if self.app_state.modality.selected_class is None:
            return {}
        return self.app_state.modality.selected_class.get_contract()

    def validate_required_instruments(self) -> tuple[bool, list[type[BaseInstrument]]]:
        if self.app_state.modality.selected_class is None:
            missing: list[type[BaseInstrument]] = []
            self.requirements_changed.emit(True, [])
            return True, missing

        missing: list[type[BaseInstrument]] = []
        for required_cls in self.app_state.modality.selected_class.REQUIRED_INSTRUMENTS:
            instances = self.instrument_service.get_instances_by_class(required_cls)
            if not instances:
                missing.append(required_cls)

        ok = len(missing) == 0
        self.requirements_changed.emit(ok, [cls.__name__ for cls in missing])
        if self.app_state.modality.running and not ok:
            self.acq_error.emit("required instrument removed during acquisition")
            self.stop()
        return ok, missing

    def configure(self, raw_params: dict[str, Any]) -> None:
        if self.app_state.modality.selected_class is None:
            raise RuntimeError("no modality selected")

        ok, missing = self.validate_required_instruments()
        if not ok:
            missing_names = ", ".join(cls.__name__ for cls in missing)
            msg = f"missing required instruments: {missing_names}"
            self.acq_error.emit(msg)
            raise RuntimeError(msg)

        if self.app_state.modality.instance is None:
            self.app_state.modality.instance = self.app_state.modality.selected_class()

        try:
            cleaned_params = coerce_parameter_values(self.app_state.modality.selected_class.PARAMETERS, raw_params)
        except Exception as exc:
            self.acq_error.emit(str(exc))
            raise

        bound: dict[type[BaseInstrument], BaseInstrument] = {}
        for required_cls in self.app_state.modality.selected_class.REQUIRED_INSTRUMENTS:
            bound_instances = self.instrument_service.get_instances_by_class(required_cls)
            if not bound_instances:
                raise RuntimeError(f"required instrument {required_cls.__name__} disappeared before configure")
            bound[required_cls] = bound_instances[0]
        for optional_cls in self.app_state.modality.selected_class.OPTIONAL_INSTRUMENTS:
            optional_instances = self.instrument_service.get_instances_by_class(optional_cls)
            if optional_instances:
                bound[optional_cls] = optional_instances[0]

        allowed_types = tuple(self.app_state.modality.selected_class.ALLOWED_OPTOCONTROLS)
        bound_opto: list[BaseOptoControl] = []
        for control in self.app_state.optocontrols:
            if not control.enabled:
                continue
            if not allowed_types:
                continue
            if not isinstance(control, allowed_types):
                continue
            control.prepare_for_acquisition()
            bound_opto.append(control)

        self.app_state.modality.instance.configure(cleaned_params, bound, bound_opto)
        self.set_parameter_values(cleaned_params)

    def start(self, continuous: bool = False) -> None:
        with self.acquisition_lock:
            instance, contract, frame_limit = self._prepare_acquisition_start(continuous=continuous)
            self.acquisition_stop_requested = threading.Event()
            self._frames_emitted = 0
            self._active_frame_limit = frame_limit
            self._active_output_contract = contract
            self._active_modality_instance = instance
            try:
                instance.prepare_acquisition_storage(frame_limit=frame_limit)
                self.app_state.modality.running = True
                self.acq_started.emit()
                self.acquisition_thread = instance.run_acquisition_threaded(
                    on_frame=self.handle_frame,
                    frame_limit=frame_limit,
                    should_stop=self.acquisition_stop_requested.is_set,
                    on_error=self.handle_acquisition_error,
                    on_finished=self.handle_acquisition_finished,
                )
            except Exception:
                self.app_state.modality.running = False
                self._active_modality_instance = None
                self._active_output_contract = ""
                raise

    def handle_frame(self, data: np.ndarray) -> None:
        if not isinstance(data, np.ndarray):
            raise TypeError(f"modality returned {type(data).__name__}, expected numpy.ndarray")

        contract = self._active_output_contract
        if contract and not matches_array_contract(data, contract):
            raise ValueError(
                f"modality returned payload not matching contract '{contract}', "
                f"shape={data.shape}, dtype={data.dtype}"
            )

        instance = self._require_active_instance()
        instance.save_acquired_frame(data, frame_index=self._frames_emitted)
        self._frames_emitted += 1
        self.data_ready.emit(data)

    def handle_acquisition_error(self, error: Exception) -> None:
        self.acq_error.emit(str(error))

    def handle_acquisition_finished(self, frame_count: int, error: Exception | None) -> None:
        try:
            self.app_state.modality.running = False
        finally:
            instance = self._active_modality_instance
            if instance is not None:
                try:
                    instance.finalize_acquisition_storage(
                        frame_count=frame_count,
                        frame_limit=self._active_frame_limit,
                        error=error,
                    )
                except Exception as finalize_exc:
                    self.acq_error.emit(str(finalize_exc))
            self.acq_stopped.emit()
            self.acquisition_thread = None
            self._active_modality_instance = None
            self._active_output_contract = ""

    def stop(self) -> None:
        self.acquisition_stop_requested.set()
        instance = self._active_modality_instance or self.app_state.modality.instance
        if instance is None:
            return
        try:
            if self.app_state.modality.running:
                instance.stop()
        except Exception as exc:
            self.acq_error.emit(str(exc))
            raise

    def set_parameter_values(self, raw_params: dict[str, Any]) -> None:
        self.app_state.modality.configured_params = [ParameterValue(label=k, value=v) for k, v in raw_params.items()]
        self.modality_params_changed.emit(dict(raw_params))

    def get_parameter_values(self) -> dict[str, Any]:
        return {entry.label: entry.value for entry in self.app_state.modality.configured_params}

    def _prepare_acquisition_start(self, *, continuous: bool) -> tuple[BaseModality, str, int | None]:
        selected_class = self.app_state.modality.selected_class
        instance = self.app_state.modality.instance
        if selected_class is None or instance is None:
            raise RuntimeError("no modality selected")
        if not self.app_state.modality.configured_params:
            raise RuntimeError("modality is not configured")
        if self.app_state.modality.running:
            raise RuntimeError("acquisition already running")

        ok, missing = self.validate_required_instruments()
        if not ok:
            missing_names = ", ".join(cls.__name__ for cls in missing)
            msg = f"cannot start acquisition, missing required instruments: {missing_names}"
            self.acq_error.emit(msg)
            raise RuntimeError(msg)

        frame_limit = None if continuous else instance.get_frame_limit()
        contract = str(getattr(selected_class, "OUTPUT_DATA_CONTRACT", ""))
        return instance, contract, frame_limit

    def _require_active_instance(self) -> BaseModality:
        instance = self._active_modality_instance
        if instance is None:
            raise RuntimeError("acquisition callback received without active modality instance")
        return instance



from datetime import datetime
from matplotlib import pyplot as plt
from matplotlib.gridspec import GridSpec
import numpy as np
import os
from scipy.optimize import curve_fit
from time import sleep
from typing import Tuple

from Swabian.TimeTagger import scanTimeTagger, createTimeTagger, Countrate, getVersion, StartStop

# SETTINGS
# Please change the settings according to your setup

channels = [1, 2]  # Use negative numbers to trigger on falling edge
min_trigger_voltage = 0.1  # V
max_trigger_voltage = 1.0  # V
data_points = 11  # per scan
analysis_range = 100000  # +/- .. ps around the maximum of the StartStop measurement
warmup_time = 0  # s
integration_time = 1  # s
jitter_analysis = False  # requires periodic signals, set to False otherwise

print("Time Tagger - Trigger Level Scan Script\n")
print("Settings (can be changed within the script)")
print(f"^ Channels: {channels} (positive numbers: rising edge, negative numbers: falling edge)")
print(f"^ Trigger level scan range from {min_trigger_voltage*1e3} mV to {max_trigger_voltage*1e3} in {data_points} steps.")
if jitter_analysis:
    print("^ Jitter analysis for periodic signals is enabled.")
else:
    print("^ Jitter analysis for periodic signals is disabled.")
print()


def gaussian(x: np.ndarray, a: float, mu: float, sigma: float) -> np.ndarray:
    return a * np.exp(-0.5 * ((x - mu) / sigma)**2)


def calc_mean_and_jitter(x: np.ndarray, y: np.ndarray) -> Tuple[float, float, float, float, float]:
    N = np.sum(y)
    if N == 0:
        return np.nan, np.nan, np.nan, np.nan, np.nan
    sum_z = np.sum(y * x)
    z_mean = sum_z / N
    variance = np.sum(y * (x - z_mean) ** 2) / N
    rms = np.sqrt(variance)
    x0 = x[y.argmax()]
    y0 = y.max()
    try:
        pop, _ = curve_fit(gaussian, x, y, p0=(y0, x0, rms))
    except RuntimeError:
        return z_mean, rms, np.nan, np.nan, np.nan
    a, mu, sigma = pop
    return z_mean, rms, a, mu, sigma


# Create Trigger level array as specified in the settings
trigger_levels = np.linspace(min_trigger_voltage, max_trigger_voltage, num=data_points)

# Create output folder
output_folder = 'trigger_level_scan/'
if not os.path.exists(output_folder):
    os.mkdir(output_folder)

# Initialize TimeTagger
print("Time Tagger Version: {}".format(getVersion()))
taggers = scanTimeTagger()
assert len(taggers) == 1, "Exactly one Time Tagger must be attached which is not in use."
tagger = createTimeTagger()
print("Time Tagger: {}  SN: {}  t = {}".format(tagger.getModel(), tagger.getSerial(), datetime.now()))

# Countrate measurement is used to analyze the count rates of all channels
countrate = Countrate(tagger, channels)
# StartStop measurements are used to analyze the autocorrelation jitter for periodic signals
start_stop_measurement = []
for channel in channels:
    start_stop_measurement.append(StartStop(tagger, channel, channel, binwidth=1))

if warmup_time == 0:
    print('For jitter analysis: please consider a warmup time of 60s for the final measurement.')
else:
    print('Wait {}s to warmup...'.format(warmup_time))
    sleep(warmup_time)
print()

# preallocate the result arrays
result = np.zeros([len(trigger_levels), len(channels)])
result_autocorr_rms = np.full_like(result, np.nan)
result_autocorr_fit = np.full_like(result, np.nan)

# Setup plots
plt.ion()
fig = plt.figure(figsize=(16, 10))
colors = plt.get_cmap('tab20').colors
gs = GridSpec(2, 2, figure=fig)
ax1 = fig.add_subplot(gs[:, 0])
ax2 = fig.add_subplot(gs[0, 1])
ax3 = fig.add_subplot(gs[1:, 1])

# Start measurement and analysis
for i, triggerlevel in enumerate(trigger_levels):
    # Measure input signals
    for channel in channels:
        tagger.setTriggerLevel(channel, triggerlevel)

    countrate.startFor(integration_time * 1e12)
    for start_stop in start_stop_measurement:
        start_stop.startFor(integration_time * 1e12)

    countrate.waitUntilFinished()
    for start_stop in start_stop_measurement:
        start_stop.waitUntilFinished()
    result[i, :] = countrate.getData()

    # Plot and analyze autocorrelation
    ax1.clear()
    for i_ss, (channel, start_stop) in enumerate(zip(channels, start_stop_measurement)):
        autocorr = start_stop.getData()
        if autocorr.size == 0:
            continue
        max_t = autocorr[autocorr[:, 1].argmax(), 0]
        analysis_dt = (int(max_t - analysis_range), int(max_t + analysis_range))
        analysis_index = np.where(
            np.logical_and(analysis_dt[0] <= autocorr[:, 0], autocorr[:, 0] < analysis_dt[1]))[0]
        if len(analysis_index) > 1:
            x = autocorr[analysis_index, 0]
            counts = autocorr[analysis_index, 1]
            ax1.plot(x, counts, color=colors[2 * i_ss + 1], label=f'data ch {channel}')
            if jitter_analysis:
                rms_x0, rms_jitter, fit_a, fit_mu, fit_jitter = calc_mean_and_jitter(x, counts)
                ax1.plot(x, gaussian(x, fit_a, fit_mu, fit_jitter), ':',
                         color=colors[2 * i_ss], label=f'fit rms {fit_jitter:.1f} ps, ch {channel}')
                result_autocorr_rms[i, i_ss] = rms_jitter
                result_autocorr_fit[i, i_ss] = fit_jitter
        else:
            x = autocorr[:, 0]
            counts = autocorr[:, 1]
            ax1.plot(x, counts, color=colors[i], label=f'data ch {channel}')
    ax1.set_title(f"Histogram channel {channel}, triggerlevel {triggerlevel*1e3:.0f} mV")
    ax1.set_xlabel("Time Difference (ps)")
    ax1.set_ylabel("Counts")
    ax1.legend()

    # Plot countrate
    ax2.clear()
    if np.all(result <= 0):
        ax2.plot(trigger_levels * 1e3, result, label=str(channels))
    else:
        ax2.semilogy(trigger_levels * 1e3, result, label=str(channels))
    ax2.set_title("Countrate vs Trigger Level")
    ax2.legend([f"ch {c}" for c in channels])
    ax2.set_xlim(min_trigger_voltage * 1e3, max_trigger_voltage * 1e3)
    ax2.set_xlabel("Trigger Level (mV)")
    ax2.set_ylabel("Count rate (Hz)")

    # Plot jitter for periodic signals
    if jitter_analysis:
        ax3.clear()
        ax3.set_prop_cycle(color=colors[0::2])
        ax3.semilogy(trigger_levels * 1e3, result_autocorr_rms, label=str(channels))
        ax3.set_prop_cycle(color=colors[1::2])
        ax3.semilogy(trigger_levels * 1e3, result_autocorr_fit, ':', label=str(channels))
        ax3.set_ylabel("Single Channel jitter * sqrt(2) (ps)")
        ax3.legend([f"ch {c} rms" for c in channels] + [f"ch {c} fitted" for c in channels])
        ax3.set_xlim(min_trigger_voltage * 1e3, max_trigger_voltage * 1e3)
        ax3.set_title("Jitter vs Trigger Level")
        ax3.set_xlabel("Trigger Level (mV)")

    plt.pause(0.1)
    fig.tight_layout()
    fig.savefig(output_folder + f'trigger_level_scan_countrate_jitter_{int(triggerlevel*1e3)}_mV.png')

    for i_channel, channel in enumerate(channels):
        msg = (
            f"Trigger level ch {channel}: {triggerlevel*1e3:4.0f} mV, countrate {result[i, i_channel]/1e3:8.3f} kHz, "
            f"rms jit {result_autocorr_rms[i, i_channel]:5.1f} ps, "
            f"fitted jit {result_autocorr_fit[i, i_channel]:5.1f} ps  ({i+1}/{len(trigger_levels)})"
        )
        print(msg)

del tagger
plt.ioff()
plt.show()



from __future__ import annotations

from PyQt6.QtWidgets import QWidget

from pyrpoc.gui.main_widgets.acquisition_mgr.handlers import (
    handle_modality_selected,
    on_parameter_values_changed,
    handle_requirements_changed,
    on_continuous_clicked,
    on_modality_selected,
    on_parameter_widgets_changed,
    on_service_error,
    on_start_clicked,
    on_stop_clicked,
    populate_modalities,
)
from pyrpoc.gui.main_widgets.acquisition_mgr.state import AcquisitionManagerState
from pyrpoc.gui.main_widgets.acquisition_mgr.ui import build_acquisition_manager_ui
from pyrpoc.services.modality_service import ModalityService


class AcquisitionManagerWidget(QWidget):
    def __init__(self, modality_service: ModalityService, parent: QWidget | None = None):
        super().__init__(parent)
        self.modality_service = modality_service
        self.state = AcquisitionManagerState()
        self.ui = build_acquisition_manager_ui(self)
        # Compatibility aliases for existing callers.
        self.modality_combo = self.ui.modality_combo
        self.refresh_btn = self.ui.refresh_btn
        self.start_btn = self.ui.start_btn
        self.continuous_btn = self.ui.continuous_btn
        self.stop_btn = self.ui.stop_btn
        self.status_label = self.ui.status_label
        self.params_container = self.ui.params_container
        self.params_layout = self.ui.params_layout
        self.param_widgets = self.state.param_widgets
        self.param_defs = self.state.param_defs

        self._wire_signals()
        self._set_acquiring_ui(False)
        self._populate_modalities()

    def _wire_signals(self) -> None:
        self.refresh_btn.clicked.connect(self._populate_modalities)
        self.modality_combo.currentTextChanged.connect(self._on_modality_selected)
        self.start_btn.clicked.connect(self._on_start_clicked)
        self.continuous_btn.clicked.connect(self._on_continuous_clicked)
        self.stop_btn.clicked.connect(self._on_stop_clicked)

        self.modality_service.modality_selected.connect(self._handle_modality_selected)
        self.modality_service.modality_params_changed.connect(self._on_parameter_values_changed)
        self.modality_service.requirements_changed.connect(self._handle_requirements_changed)
        self.modality_service.acq_started.connect(self._on_acq_started)
        self.modality_service.acq_stopped.connect(self._on_acq_stopped)
        self.modality_service.acq_error.connect(self._on_service_error)

    def _populate_modalities(self) -> None:
        populate_modalities(self)

    def _on_modality_selected(self, key: str) -> None:
        on_modality_selected(self, key)

    def _handle_modality_selected(self, key: str) -> None:
        handle_modality_selected(self, key)

    def _on_start_clicked(self) -> None:
        on_start_clicked(self)

    def _on_continuous_clicked(self) -> None:
        on_continuous_clicked(self)

    def _on_stop_clicked(self) -> None:
        on_stop_clicked(self)

    def _handle_requirements_changed(self, ok: bool, missing_names: list[str]) -> None:
        handle_requirements_changed(self, ok, missing_names)

    def _on_service_error(self, message: str) -> None:
        on_service_error(self, message)
        self._set_acquiring_ui(False)

    def _on_parameter_widgets_changed(self) -> None:
        on_parameter_widgets_changed(self)

    def _on_parameter_values_changed(self, values: object) -> None:
        on_parameter_values_changed(self, values)

    def _on_acq_started(self) -> None:
        self.status_label.setText("Status: acquiring")
        self._set_acquiring_ui(True)

    def _on_acq_stopped(self) -> None:
        self.status_label.setText("Status: stopped")
        self._set_acquiring_ui(False)

    def _set_acquiring_ui(self, acquiring: bool) -> None:
        self.start_btn.setEnabled(not acquiring)
        self.continuous_btn.setEnabled(not acquiring)
        self.stop_btn.setEnabled(acquiring)



from __future__ import annotations

from typing import Any


class Registry:
    def __init__(self, name: str, base_class: type):
        self.name = name
        self.base_class = base_class
        self.entries: dict[str, type] = {}

    def register(self, key: str):
        def decorator(cls: type):
            if not issubclass(cls, self.base_class):
                raise TypeError(f"{cls.__name__} must inherit from {self.base_class.__name__}")
            if key in self.entries:
                raise KeyError(f"{key!r} is already registered in {self.name}")
            self.entries[key] = cls
            return cls

        return decorator

    def list_keys(self) -> list[str]:
        return sorted(self.entries.keys())

    def get_registered(self) -> list[str]:
        return self.list_keys()

    def get_class(self, key: str) -> type:
        if key not in self.entries:
            raise KeyError(f"{key!r} is not registered in {self.name}")
        return self.entries[key]

    def create(self, key: str, **kwargs: Any) -> Any:
        cls = self.get_class(key)
        return cls(**kwargs)

    def describe(self, key: str) -> dict[str, Any]:
        cls = self.get_class(key)
        return {
            "key": key,
            "class_name": cls.__name__,
            "display_name": getattr(cls, "DISPLAY_NAME", cls.__name__),
            "parameters": getattr(cls, "PARAMETERS", {}),
            "config_parameters": getattr(cls, "CONFIG_PARAMETERS", {}),
            "display_parameters": getattr(cls, "DISPLAY_PARAMETERS", {}),
            "actions": getattr(cls, "ACTIONS", []),
            "contract": cls.get_contract() if hasattr(cls, "get_contract") else {},
        }

    def describe_all(self) -> list[dict[str, Any]]:
        return [self.describe(key) for key in self.list_keys()]



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any

from PyQt6.QtWidgets import QWidget

from pyrpoc.backend_utils.state_helpers import export_object_state, import_object_state, make_instance_id


class BaseInstrument(ABC):
    INSTRUMENT_KEY: str = "base_instrument"
    DISPLAY_NAME: str = "Base Instrument"
    PERSISTENCE_FIELDS: tuple[str, ...] | None = None
    PERSISTENCE_EXCLUDE_FIELDS: tuple[str, ...] = (
        "alias",
        "connected",
        "instance_id",
        "last_error",
        "user_label",
        "widget",
    )

    def __init__(
        self,
        alias: str | None = None,
        *,
        instance_id: str | None = None,
        user_label: str | None = None,
        connected: bool = False,
    ):
        self.alias = alias or self.INSTRUMENT_KEY
        self.instance_id = instance_id or make_instance_id(self.alias)
        self.user_label = user_label
        self.connected = bool(connected)
        self.last_error: str | None = None

    @property
    def type_key(self) -> str:
        """Registry key used by persistence and inventory rows to recreate this instance."""
        return self.alias

    @classmethod
    def get_contract(cls) -> dict[str, Any]:
        return {
            "instrument_key": cls.INSTRUMENT_KEY,
            "display_name": cls.DISPLAY_NAME,
        }

    @classmethod
    def get_widget_contract(cls) -> dict[str, Any]:
        """Return only UI-related metadata for discovery-driven widgets.

        Call flow:
        - UI code queries `instrument_service.list_available()`
        - `Registry.describe()` returns `get_widget_contract()` in addition to `get_contract()`
        - Managers can use this to skip special-case imports while deciding how to render.
        """
        return {}

    @abstractmethod
    def get_widget(
        self,
        parent: QWidget | None = None,
        on_change: Callable[[], None] | None = None,
    ) -> "BaseInstrumentWidget":
        """Create or reuse a concrete widget for this instance.

        Call flow:
        - user expands an instance card in InstrumentManager
        - instrument_mgr.handlers.on_expand_requested asks InstrumentService for the widget
        - InstrumentService delegates to this method to keep all concrete imports localized.
        """
        raise NotImplementedError

    def prepare_for_acquisition(self) -> tuple[Any, ...]:
        """Return a compact acquisition payload for future modality hooks.

        Call flow:
        - modality code may call this after requirements validation
        - default payload includes only stable identifiers to keep phase-1 minimal.
        """
        return (self.alias,)

    def get_collapsed_summary(self) -> str:
        """Return short text shown next to instrument name in collapsed manager cards."""
        return ""

    def export_persistence_state(self) -> dict[str, Any]:
        return export_object_state(
            self,
            include_fields=self.PERSISTENCE_FIELDS,
            exclude_fields=self.PERSISTENCE_EXCLUDE_FIELDS,
        )

    def import_persistence_state(self, state: dict[str, Any]) -> None:
        import_object_state(
            self,
            state,
            include_fields=self.PERSISTENCE_FIELDS,
            exclude_fields=self.PERSISTENCE_EXCLUDE_FIELDS,
        )

    def connect(self) -> bool:
        return True


class BaseInstrumentWidget(QWidget):
    """Base Qt widget for a single instrument instance.

    Call flow:
    - user clicks Expand in InstrumentManager card
    - instrument_mgr.handlers.on_expand_requested -> InstrumentService.get_widget
    - that call returns a `BaseInstrumentWidget` from concrete `BaseInstrument.get_widget`
    """

    def __init__(self, instrument: BaseInstrument, on_change: Callable[[], None] | None = None, parent: QWidget | None = None):
        super().__init__(parent)
        self.instrument = instrument
        self.on_change = on_change

    def refresh_from_model(self) -> None:
        """Load current instrument-level model state after restore/rebuild."""
        return None

    def _request_model_persist(self) -> None:
        if self.on_change is not None:
            self.on_change()



from __future__ import annotations

from collections.abc import Callable

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QDoubleSpinBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QSpinBox,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from pyrpoc.instruments.base_instrument import BaseInstrumentWidget

if False:  # pragma: no cover
    from pyrpoc.instruments.confocal_daq import ConfocalDAQInstrument


class ConfocalDAQInstrumentWidget(BaseInstrumentWidget):
    def __init__(
        self,
        instrument: "ConfocalDAQInstrument",
        on_change: Callable[[], None] | None = None,
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(instrument, on_change=on_change, parent=parent)
        self.instrument = instrument
        self._channel_leds: list[QToolButton] = []
        self._building_leds = False

        root = QVBoxLayout(self)
        root.setContentsMargins(4, 4, 4, 4)
        root.setSpacing(8)

        grid_box = QGroupBox("Confocal DAQ Configuration", self)
        grid = QGridLayout(grid_box)
        grid.setContentsMargins(8, 8, 8, 8)
        grid.setHorizontalSpacing(8)
        grid.setVerticalSpacing(6)

        self.device_edit = QLineEdit(self.instrument.device_name, self)
        self.device_edit.editingFinished.connect(self._on_core_params_changed)
        grid.addWidget(QLabel("DAQ Device", self), 0, 0)
        grid.addWidget(self.device_edit, 0, 1)

        self.sample_rate_spin = QDoubleSpinBox(self)
        self.sample_rate_spin.setDecimals(1)
        self.sample_rate_spin.setRange(1.0, 5_000_000.0)
        self.sample_rate_spin.setSingleStep(1_000.0)
        self.sample_rate_spin.setValue(float(self.instrument.sample_rate_hz))
        self.sample_rate_spin.valueChanged.connect(self._on_core_params_changed)
        grid.addWidget(QLabel("Sample Rate (Hz)", self), 1, 0)
        grid.addWidget(self.sample_rate_spin, 1, 1)

        self.fast_ao_spin = QSpinBox(self)
        self.fast_ao_spin.setRange(0, 31)
        self.fast_ao_spin.setValue(int(self.instrument.fast_axis_ao))
        self.fast_ao_spin.valueChanged.connect(self._on_core_params_changed)
        grid.addWidget(QLabel("Fast Axis AO", self), 2, 0)
        grid.addWidget(self.fast_ao_spin, 2, 1)

        self.slow_ao_spin = QSpinBox(self)
        self.slow_ao_spin.setRange(0, 31)
        self.slow_ao_spin.setValue(int(self.instrument.slow_axis_ao))
        self.slow_ao_spin.valueChanged.connect(self._on_core_params_changed)
        grid.addWidget(QLabel("Slow Axis AO", self), 3, 0)
        grid.addWidget(self.slow_ao_spin, 3, 1)

        root.addWidget(grid_box)

        led_box = QGroupBox("Active AI Channels", self)
        led_layout = QVBoxLayout(led_box)
        led_layout.setContentsMargins(8, 8, 8, 8)
        led_layout.setSpacing(6)
        led_layout.addWidget(QLabel("Click to toggle channel activity", self))

        self.led_row = QHBoxLayout()
        self.led_row.setContentsMargins(0, 0, 0, 0)
        self.led_row.setSpacing(6)
        self.led_row.setAlignment(Qt.AlignmentFlag.AlignLeft)
        led_layout.addLayout(self.led_row)
        root.addWidget(led_box)

        self._rebuild_led_row()

    def refresh_from_model(self) -> None:
        self.device_edit.setText(self.instrument.device_name)
        self.sample_rate_spin.setValue(float(self.instrument.sample_rate_hz))
        self.fast_ao_spin.setValue(int(self.instrument.fast_axis_ao))
        self.slow_ao_spin.setValue(int(self.instrument.slow_axis_ao))
        self._rebuild_led_row()

    def _on_core_params_changed(self, _value: float | int = 0) -> None:
        del _value
        self.instrument.device_name = self.device_edit.text().strip() or "Dev1"
        self.instrument.sample_rate_hz = float(self.sample_rate_spin.value())
        self.instrument.fast_axis_ao = int(self.fast_ao_spin.value())
        self.instrument.slow_axis_ao = int(self.slow_ao_spin.value())
        self._request_model_persist()

    def _on_led_toggled(self, index: int, checked: bool) -> None:
        if self._building_leds:
            return
        if 0 <= index < len(self.instrument.active_ai_channels):
            self.instrument.active_ai_channels[index] = bool(checked)
            self._request_model_persist()

    def _rebuild_led_row(self) -> None:
        self._building_leds = True
        while self.led_row.count() > 0:
            item = self.led_row.takeAt(0)
            child = item.widget()
            if child is not None:
                child.setParent(None)
                child.deleteLater()
        self._channel_leds.clear()

        for i, ai_idx in enumerate(self.instrument.ai_channel_numbers):
            led = QToolButton(self)
            led.setCheckable(True)
            led.setChecked(bool(self.instrument.active_ai_channels[i]))
            led.setText(f"AI{ai_idx}")
            led.setToolTip(f"Toggle AI{ai_idx}")
            led.setStyleSheet(
                "QToolButton {"
                "padding: 4px 8px;"
                "border: 1px solid palette(mid);"
                "border-radius: 10px;"
                "background: palette(base);"
                "}"
                "QToolButton:checked {"
                "background: palette(highlight);"
                "color: palette(highlighted-text);"
                "border: 1px solid palette(highlight);"
                "font-weight: 700;"
                "}"
            )
            led.toggled.connect(lambda checked, idx=i: self._on_led_toggled(idx, checked))
            self.led_row.addWidget(led)
            self._channel_leds.append(led)

        self._building_leds = False



from __future__ import annotations

from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtGui import QColor
from PyQt6.QtWidgets import (
    QCheckBox,
    QFrame,
    QHBoxLayout,
    QLabel,
    QToolButton,
    QVBoxLayout,
    QWidget,
)


class InstanceCardWidget(QFrame):
    expand_requested = pyqtSignal(object)
    enable_toggled = pyqtSignal(object, bool)
    remove_requested = pyqtSignal(object)

    def __init__(self, state_obj: object, title: str, parent: QWidget | None = None):
        super().__init__(parent)
        self.state_obj = state_obj
        self._expanded = False
        self._enable_guard = False

        self.setObjectName("optoControlCard")
        self.setProperty("enabledCard", False)
        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setStyleSheet(self._compose_stylesheet(enabled=False))

        root = QVBoxLayout(self)
        root.setContentsMargins(6, 4, 6, 4)
        root.setSpacing(2)

        header_row = QHBoxLayout()
        header_row.setContentsMargins(0, 0, 0, 0)
        header_row.setSpacing(4)

        self.expand_btn = QToolButton(self)
        self.expand_btn.setArrowType(Qt.ArrowType.RightArrow)
        self.expand_btn.setAutoRaise(True)
        self.expand_btn.setToolTip("Expand")
        self.expand_btn.clicked.connect(self._on_expand_clicked)
        header_row.addWidget(self.expand_btn)

        self.enable_checkbox = QCheckBox("", self)
        self.enable_checkbox.setToolTip("Enable")
        self.enable_checkbox.toggled.connect(self._on_enable_toggled)
        header_row.addWidget(self.enable_checkbox)

        self.title_label = QLabel(title, self)
        header_row.addWidget(self.title_label, 1)

        self.remove_btn = QToolButton(self)
        self.remove_btn.setAutoRaise(True)
        self.remove_btn.setText("X")
        self.remove_btn.setStyleSheet("QToolButton { color: palette(text); font-weight: 700; }")
        self.remove_btn.setToolTip("Remove")
        self.remove_btn.clicked.connect(lambda: self.remove_requested.emit(self.state_obj))
        header_row.addWidget(self.remove_btn)

        root.addLayout(header_row)

        self.body_container = QWidget(self)
        self.body_layout = QVBoxLayout(self.body_container)
        self.body_layout.setContentsMargins(0, 2, 0, 0)
        self.body_layout.setSpacing(4)
        self.body_container.setVisible(False)
        root.addWidget(self.body_container)

    def _on_expand_clicked(self) -> None:
        self.expand_requested.emit(self.state_obj)

    def _on_enable_toggled(self, checked: bool) -> None:
        self._apply_enabled_visual(bool(checked))
        if self._enable_guard:
            return
        self.enable_toggled.emit(self.state_obj, checked)

    def set_expanded(self, expanded: bool) -> None:
        self._expanded = expanded
        self.body_container.setVisible(expanded)
        self.expand_btn.setArrowType(Qt.ArrowType.DownArrow if expanded else Qt.ArrowType.RightArrow)
        self.expand_btn.setToolTip("Collapse" if expanded else "Expand")

    def is_expanded(self) -> bool:
        '''Report current expanded state for expand toggle handlers in manager logic.'''
        return self._expanded

    def set_marker_text(self, text: str) -> None:
        del text

    def set_local_status(self, text: str) -> None:
        self.setToolTip(text)

    def set_enable_checked(self, checked: bool, guarded: bool = True) -> None:
        if guarded:
            self._enable_guard = True
        self.enable_checkbox.setChecked(checked)
        self._apply_enabled_visual(bool(checked))
        if guarded:
            self._enable_guard = False

    def set_enable_visible(self, visible: bool) -> None:
        self.enable_checkbox.setVisible(visible)

    def set_body_widget(self, body: QWidget | None) -> None:
        while self.body_layout.count():
            item = self.body_layout.takeAt(0)
            child = item.widget()
            if child is not None:
                child.setParent(None)
                child.deleteLater()
        if body is not None:
            self.body_layout.addWidget(body)

    def _apply_enabled_visual(self, enabled: bool) -> None:
        self.setProperty("enabledCard", bool(enabled))
        self.setStyleSheet(self._compose_stylesheet(enabled=bool(enabled)))
        self.style().unpolish(self)
        self.style().polish(self)
        self.update()

    def _compose_stylesheet(self, enabled: bool) -> str:
        highlight = QColor(self.palette().color(self.palette().ColorRole.Highlight))
        muted_bg = QColor(highlight)
        muted_bg.setAlpha(46)
        muted_border = QColor(highlight)
        muted_border.setAlpha(150)
        enabled_block = ""
        if enabled:
            enabled_block = (
                "#optoControlCard {"
                f"background: rgba({muted_bg.red()}, {muted_bg.green()}, {muted_bg.blue()}, {muted_bg.alpha()});"
                f"border: 1px solid rgba({muted_border.red()}, {muted_border.green()}, {muted_border.blue()}, {muted_border.alpha()});"
                "}"
            )
        return (
            "#optoControlCard {"
            "background: palette(base);"
            "border: 1px solid palette(midlight);"
            "border-radius: 6px;"
            "}"
            "#optoControlCard > QWidget, #optoControlCard > QWidget QWidget {"
            "background: transparent;"
            "}"
            "#optoControlCard QLabel, #optoControlCard QCheckBox, #optoControlCard QToolButton {"
            "background: transparent;"
            "}"
            "#optoControlCard QCheckBox::indicator:checked {"
            "background: palette(highlight);"
            "border: 1px solid palette(highlight);"
            "color: palette(highlighted-text);"
            "}"
            "#optoControlCard QCheckBox::indicator:unchecked {"
            "background: transparent;"
            "border: 1px solid palette(mid);"
            "}"
            f"{enabled_block}"
        )



from __future__ import annotations

from datetime import datetime, timezone
import json
from pathlib import Path
from typing import Any

import numpy as np
import tifffile

from pyrpoc.backend_utils.array_contracts import CONTRACT_CHW_FLOAT32
from pyrpoc.instruments.confocal_daq import ConfocalDAQInstrument
from pyrpoc.backend_utils.opto_control_contexts import MaskContext
from pyrpoc.optocontrols.base_optocontrol import BaseOptoControl
from pyrpoc.optocontrols.mask import MaskOptoControl
from pyrpoc.backend_utils.parameter_utils import (
    CheckboxParameter,
    NumberParameter,
    PathParameter,
)

from .acquisition_functions.daq_helpers import (
    DaqUnavailableError,
    acquire_daq_split_confocal,
    extract_mask_contexts,
)
from .acquisition_functions.toy_data import generate_toy_split_confocal_frame
from .base_modality import BaseModality
from .mod_registry import modality_registry


@modality_registry.register("split_confocal")
class SplitConfocalModality(BaseModality):
    MODALITY_KEY = "split_confocal"
    DISPLAY_NAME = "Split Confocal"
    PARAMETERS = {
        "scan": [
            NumberParameter(
                label="X Pixels",
                default=512,
                minimum=8,
                tooltip="Number of pixels in X",
                number_type=int,
            ),
            NumberParameter(
                label="Y Pixels",
                default=512,
                minimum=8,
                tooltip="Number of pixels in Y",
                number_type=int,
            ),
            NumberParameter(
                label="Extra Steps Left",
                default=300,
                minimum=0,
                tooltip="Extra scan steps at left edge (stored only for now)",
                number_type=int,
            ),
            NumberParameter(
                label="Extra Steps Right",
                default=20,
                minimum=0,
                tooltip="Extra scan steps at right edge (stored only for now)",
                number_type=int,
            ),
            NumberParameter(
                label="Fast Axis Offset",
                default=0.0,
                tooltip="Fast-axis offset",
                number_type=float,
            ),
            NumberParameter(
                label="Fast Axis Amplitude",
                default=1.0,
                minimum=1e-6,
                tooltip="Fast-axis amplitude",
                number_type=float,
            ),
            NumberParameter(
                label="Slow Axis Offset",
                default=0.0,
                tooltip="Slow-axis offset",
                number_type=float,
            ),
            NumberParameter(
                label="Slow Axis Amplitude",
                default=1.0,
                minimum=1e-6,
                tooltip="Slow-axis amplitude",
                number_type=float,
            ),
            NumberParameter(
                label="Dwell Time (us)",
                default=2.0,
                minimum=0.1,
                tooltip="Pixel dwell time",
                number_type=float,
            ),
        ],
        "acquisition": [
            NumberParameter(
                label="t0 Samples",
                default=1,
                minimum=1,
                tooltip="Number of samples in the first subpixel window",
                number_type=int,
            ),
            NumberParameter(
                label="t1 Samples",
                default=0,
                minimum=0,
                tooltip="Number of samples to discard between t0 and t2",
                number_type=int,
            ),
            CheckboxParameter(
                label="save_enabled",
                display_label="save_enabled",
                default=False,
                required=False,
                tooltip="Enable saving frames and acquisition metadata",
            ),
            PathParameter(
                label="save_path",
                display_label="save_path",
                default="acquisition",
                required=False,
                tooltip="Base name/path for saved TIFF files (e.g. /dir/acquisition)",
            ),
            NumberParameter(
                label="num_frames",
                display_label="num_frames",
                default=1,
                required=False,
                minimum=1,
                tooltip="Number of frames to capture",
                number_type=int,
            ),
        ],
    }
    REQUIRED_INSTRUMENTS = [ConfocalDAQInstrument]
    OPTIONAL_INSTRUMENTS = []
    ALLOWED_OPTOCONTROLS = [MaskOptoControl]
    OUTPUT_DATA_CONTRACT = CONTRACT_CHW_FLOAT32
    ALLOWED_DISPLAYS = ["tiled_2d", "multichan_overlay"]

    def __init__(self):
        super().__init__()
        self._frame_idx = 0
        self._mask_contexts: list[MaskContext] = []
        self._daq_instrument: ConfocalDAQInstrument = None
        self._active_ai_channels: list[int] = []
        self._pending_auxiliary: dict[str, np.ndarray] = {}
        self._save_enabled = False
        self._save_root_path: Path | None = None
        self._save_json_path: Path | None = None
        self._save_channel_paths: dict[str, Path] = {}
        self._save_channel_labels: list[str] = []
        self._auxiliary_payload_buffers: dict[str, list[np.ndarray]] = {}
        self._auxiliary_paths: dict[str, Path] = {}
        self._saved_frame_count = 0
        self._run_id = 0
        self._run_started_at = ""
        self._run_frame_limit: int | None = 1

    def configure(
        self,
        params: dict[str, Any],
        instruments: dict[type[ConfocalDAQInstrument], ConfocalDAQInstrument],
        opto_controls: list[BaseOptoControl],
    ) -> None:
        self._params = dict(params)
        self._instruments = dict(instruments)
        instrument = self._instruments.get(ConfocalDAQInstrument)
        if instrument is None:
            raise RuntimeError("ConfocalDAQInstrument missing during configure")
        self._daq_instrument = instrument
        self._mask_contexts = extract_mask_contexts(opto_controls)
        self._active_ai_channels = self._get_active_ai_channels()
        self._frame_idx = 0
        self._pending_auxiliary = {}
        self._save_enabled = bool(self._params.get("save_enabled", False))
        self._save_root_path = self._coerce_save_root(self._params.get("save_path"))
        self._save_json_path = None
        self._save_channel_paths = {}
        self._save_channel_labels = []
        self._auxiliary_payload_buffers = {}
        self._auxiliary_paths = {}
        self._saved_frame_count = 0
        self._run_frame_limit = 1
        self._configured = True

    def start(self) -> None:
        if not self._configured:
            raise RuntimeError("modality must be configured before start")
        if self._daq_instrument is None:
            raise RuntimeError("ConfocalDAQInstrument not configured")
        self._running = True

    def acquire_once(self) -> np.ndarray:
        if not self._running:
            raise RuntimeError("modality is not running")
        if self._daq_instrument is None:
            raise RuntimeError("ConfocalDAQInstrument unavailable")

        settings = self._get_scan_settings()
        split = self._get_split_settings()

        try:
            frame, raw = acquire_daq_split_confocal(
                daq_instrument=self._daq_instrument,
                active_ai_channels=self._active_ai_channels,
                mask_contexts=self._mask_contexts,
                **settings,
                **split,
            )
        except Exception as exc:
            print(f"acquisition unavailable ({exc}), falling back to toy data")
            if not isinstance(exc, DaqUnavailableError) and not isinstance(exc, RuntimeError):
                raise
            pixel_samples = max(1, int(settings["dwell_time_us"] * 1e-6 * float(self._daq_instrument.sample_rate_hz)))
            frame, raw = generate_toy_split_confocal_frame(
                x_pixels= settings["x_pixels"],
                y_pixels= settings["y_pixels"],
                active_channels=self._active_ai_channels,
                frame_index=self._frame_idx,
                mask_contexts=self._mask_contexts,
                fast_axis_offset=settings["fast_axis_offset"],
                fast_axis_amplitude=settings["fast_axis_amplitude"],
                slow_axis_offset=settings["slow_axis_offset"],
                slow_axis_amplitude=settings["slow_axis_amplitude"],
                t0_samples=split["t0_samples"],
                t1_samples=split["t1_samples"],
                pixel_samples=pixel_samples,
            )

        self._pending_auxiliary = {"raw_pixel_stream": raw}
        self._frame_idx += 1
        return frame.astype(np.float32, copy=False)

    def stop(self) -> None:
        self._running = False

    def prepare_acquisition_storage(self, *, frame_limit: int | None) -> None:
        self._saved_frame_count = 0
        self._run_frame_limit = frame_limit
        self._run_id += 1
        self._run_started_at = datetime.now(timezone.utc).isoformat()
        self._save_channel_paths = {}
        self._save_channel_labels = []
        self._auxiliary_payload_buffers = {}
        self._auxiliary_paths = {}
        if not self._save_enabled:
            return
        if self._save_root_path is None:
            raise RuntimeError("save_path is required when save_enabled is true")
        self._save_root_path.parent.mkdir(parents=True, exist_ok=True)
        self._save_json_path = self._save_root_path.with_name(f"{self._save_root_path.name}_meta.json")
        self._write_metadata(None)

    def save_acquired_frame(self, frame: np.ndarray, *, frame_index: int) -> None:
        if not self._save_enabled:
            self._pending_auxiliary = {}
            return
        if self._save_root_path is None:
            raise RuntimeError("save_path is required when save_enabled is true")

        channel_data = self._split_channels(frame)
        if not self._save_channel_paths:
            labels = self._resolve_channel_labels(len(channel_data))
            self._save_channel_labels = labels
            self._save_channel_paths = {
                label: self._save_root_path.with_name(f"{self._save_root_path.name}_{label}.tiff")
                for label in labels
            }
            for path in self._save_channel_paths.values():
                if path.exists():
                    path.unlink()
        if len(channel_data) != len(self._save_channel_paths):
            raise ValueError("frame channel count does not match configured save layout")

        for (label, path), channel_frame in zip(self._save_channel_paths.items(), channel_data):
            del label
            with tifffile.TiffWriter(str(path), append=True) as writer:
                writer.write(channel_frame.astype(np.float32))

        self._append_auxiliary_payload()
        self._saved_frame_count = frame_index + 1
        self._write_metadata(None)

    def finalize_acquisition_storage(
        self,
        *,
        frame_count: int,
        frame_limit: int | None,
        error: Exception | None,
    ) -> None:
        self._saved_frame_count = frame_count
        self._run_frame_limit = frame_limit
        self._flush_auxiliary_payloads()
        self._write_metadata(str(error) if error is not None else None)

    def get_frame_limit(self) -> int:
        raw = self._params.get("num_frames", 1)
        limit = int(raw)
        if limit < 1:
            raise ValueError("num_frames must be >= 1")
        return limit

    def get_active_channel_labels(self) -> list[str]:
        labels: list[str] = []
        for channel in self._active_ai_channels:
            labels.append(f"ai{channel}_t0")
            labels.append(f"ai{channel}_t2")
        return labels

    def _get_active_ai_channels(self) -> list[int]:
        return [
            ai_channel
            for ai_channel, enabled in zip(
                self._daq_instrument.ai_channel_numbers if self._daq_instrument is not None else [],
                self._daq_instrument.active_ai_channels if self._daq_instrument is not None else [],
                strict=False,
            )
            if enabled
        ]

    def _get_scan_settings(self) -> dict[str, Any]:
        return {
            "x_pixels": int(self._params["X Pixels"]),
            "y_pixels": int(self._params["Y Pixels"]),
            "dwell_time_us": float(self._params["Dwell Time (us)"]),
            "extra_left": int(self._params["Extra Steps Left"]),
            "extra_right": int(self._params["Extra Steps Right"]),
            "fast_axis_offset": float(self._params["Fast Axis Offset"]),
            "fast_axis_amplitude": max(float(self._params["Fast Axis Amplitude"]), 1e-6),
            "slow_axis_offset": float(self._params["Slow Axis Offset"]),
            "slow_axis_amplitude": max(float(self._params["Slow Axis Amplitude"]), 1e-6),
        }

    def _get_split_settings(self) -> dict[str, Any]:
        t0 = int(self._params["t0 Samples"])
        t1 = int(self._params["t1 Samples"])
        if t0 < 1:
            raise ValueError("t0 Samples must be >= 1")
        if t1 < 0:
            raise ValueError("t1 Samples must be >= 0")
        return {
            "t0_samples": t0,
            "t1_samples": t1,
        }

    def _coerce_save_root(self, raw_save_path: Any) -> Path | None:
        if not self._save_enabled:
            return None
        if raw_save_path is None:
            raise ValueError("save_path is required when save_enabled is true")
        path = raw_save_path if isinstance(raw_save_path, Path) else Path(raw_save_path).expanduser()
        if path.suffix.lower() in {".tif", ".tiff"}:
            path = path.with_suffix("")
        return path

    def _resolve_channel_labels(self, channel_count: int) -> list[str]:
        active_labels = self.get_active_channel_labels()
        if active_labels and len(active_labels) == channel_count:
            return list(active_labels)
        return [f"channel_{index}" for index in range(channel_count)]

    def _split_channels(self, data: np.ndarray) -> list[np.ndarray]:
        if data.ndim == 2:
            return [data]
        if data.ndim == 3:
            return [data[index] for index in range(data.shape[0])]
        raise ValueError(f"unsupported frame dimensions {data.ndim}")

    def _append_auxiliary_payload(self) -> None:
        if not self._save_enabled or self._save_root_path is None:
            self._pending_auxiliary = {}
            return
        if not self._pending_auxiliary:
            return

        if not self._auxiliary_paths:
            labels = list(self._pending_auxiliary.keys())
            self._auxiliary_paths = {
                label: self._save_root_path.with_name(f"{self._save_root_path.name}_{label}.npz")
                for label in labels
            }
            for path in self._auxiliary_paths.values():
                if path.exists():
                    path.unlink()

        for label, payload in self._pending_auxiliary.items():
            frames = self._auxiliary_payload_buffers.setdefault(label, [])
            frames.append(np.asarray(payload, dtype=np.float32, copy=False))
        self._pending_auxiliary = {}

    def _flush_auxiliary_payloads(self) -> None:
        if not self._save_enabled or not self._auxiliary_paths:
            self._pending_auxiliary = {}
            return
        for label, path in self._auxiliary_paths.items():
            frames = self._auxiliary_payload_buffers.get(label, [])
            if not frames:
                continue
            payload = np.stack(frames, axis=0)
            np.savez_compressed(
                str(path),
                frames=payload,
                parameters=np.asarray(dict(self._params), dtype=object),
                frame_indices=np.arange(payload.shape[0], dtype=np.int32),
            )
        self._pending_auxiliary = {}

    def _write_metadata(self, last_error: str | None) -> None:
        if not self._save_enabled or self._save_json_path is None:
            return
        payload = {
            "run_id": self._run_id,
            "started": self._run_started_at,
            "modality_key": self.MODALITY_KEY,
            "save_root_path": str(self._save_root_path),
            "save_json_path": str(self._save_json_path),
            "tiff_paths": {label: str(path) for label, path in self._save_channel_paths.items()},
            "auxiliary_paths": {label: str(path) for label, path in self._auxiliary_paths.items()},
            "frames_saved": self._saved_frame_count,
            "frame_limit": self._run_frame_limit,
            "parameters": dict(self._params),
            "last_error": last_error,
        }
        self._save_json_path.write_text(json.dumps(payload, indent=2, default=str), encoding="utf-8")


SplitConfocal = SplitConfocalModality



"""
This script allows you to check your parameters.
"""

from json import load
from matplotlib import pyplot as plt
from pnr_constants import FRONT_EDGE, REAR_EDGE, LASER_CHANNEL, FOLDER, FILENAME

from Swabian import TimeTagger


with open(FOLDER + "parameters.json", "rb") as param_file:
    parameters: dict = load(param_file)

ttv = TimeTagger.createTimeTaggerVirtual(filename=FOLDER + FILENAME)

# Only relevant if Virtual Channels are use in the calibration file
for i in range(3):
    x = TimeTagger.SyntheticSingleTag(ttv)

# Setup photon number resolution channel
pnr = TimeTagger.Experimental.PhotonNumber(tagger=ttv,
                                           trigger_ch=LASER_CHANNEL,
                                           signal_start_ch=FRONT_EDGE,
                                           signal_stop_ch=REAR_EDGE,
                                           slope=parameters["slope"],
                                           x_intercepts=parameters["x_intercepts"],
                                           dead_time=100_000)

delayed_channels = list()
gated_channels = list()
histograms = list()
for channel in pnr.getChannels()[1:]:
    delayed = TimeTagger.DelayedChannel(tagger=ttv,
                                        input_channel=channel,
                                        delay=-parameters["delay"])
    gated = TimeTagger.GatedChannel(tagger=ttv,
                                    input_channel=FRONT_EDGE,
                                    gate_start_channel=delayed.getChannel(),
                                    gate_stop_channel=channel)
    histograms.append(TimeTagger.Histogram(tagger=ttv,
                                           click_channel=gated.getChannel(),
                                           start_channel=delayed.getChannel(),
                                           binwidth=1,
                                           n_bins=1000))
    delayed_channels.append(delayed)
    gated_channels.append(gated)

ttv.run()
ttv.waitUntilFinished()

for hist in histograms:
    plt.plot(hist.getIndex(), hist.getData())
plt.title("Split photon peaks - does it make sense? :)")
plt.legend(["1 photon"] + [f"{i} photons" for i in range(2,
           len(pnr.getChannels())-1)] + [f"≥{len(pnr.getChannels())-1} photons"])
plt.show()



"""In this example, we learn how to start a simple measurement. We use the Counter class to measure a
count rate trace on channels 1 and 2 while switching on the built-in test signals.
For a demonstration of Counter.getDataObject(), see example 2-E."""

from matplotlib import pyplot as plt
from time import sleep

from Swabian import TimeTagger

# Create a TimeTagger instance to control your hardware
tagger = TimeTagger.createTimeTagger()

# Create an instance of the Counter measurement class. It starts acquiring data immediately.
counter = TimeTagger.Counter(tagger=tagger, channels=[1, 2], binwidth=int(1e9), n_values=1000)

# Apply the built-in test signal (~0.8 to 0.9 MHz) to channel 1
tagger.setTestSignal(1, True)
print("Test signal on channel 1 enabled")
sleep(.5)

# Apply test signal to channel 2
tagger.setTestSignal(2, True)
print("Test signal on channel 2 enabled")

# After waiting two times for 0.5 s, the 1000 values should be filled
sleep(.5)

# Data is retrieved by calling the method "getData" on the measurement class.
data = counter.getData()

# Plot the result
plt.figure()
plt.plot(counter.getIndex()/1e12, data[0]*1e-3, label='channel 1')
plt.plot(counter.getIndex()/1e12, data[1]*1e-3, label='channel 2')
plt.xlabel('Time (s)')
plt.ylabel('Countrate (MHz)')
plt.legend()
plt.title('Time trace of the click rate on channel 1 and 2')
plt.annotate('''The built-in test signal(~ 800 to 900 kHz) is applied
first to channel 1 and 0.5 s later to channel 2.
The total delay time within the script is 1 s.
As you can see, the counts are still 0 at 0 s.
The Time Tagger and the analysis software are
running asynchronously. Therefore, the total processed
data is less than 1 s. To have an exact measurement
duration, please use startFor() instead of sleep().''',
             (0, 0),
             xytext=(100, 100),
             textcoords='offset pixels',
             arrowprops={'arrowstyle': '->'},
             va='bottom')
plt.tight_layout()
plt.show()

TimeTagger.freeTimeTagger(tagger)



from __future__ import annotations

import random

import cv2
import numpy as np
from PyQt6.QtCore import QPointF, QRectF, Qt, pyqtSignal
from PyQt6.QtGui import QColor, QFont, QImage, QPainter, QPainterPath, QPen, QPixmap
from PyQt6.QtWidgets import (
    QCheckBox,
    QDialog,
    QFileDialog,
    QGraphicsEllipseItem,
    QGraphicsPathItem,
    QGraphicsScene,
    QGraphicsTextItem,
    QGraphicsView,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QMenu,
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)
from superqt import QRangeSlider

from pyrpoc.rpoc.segmentation_methods import labels_to_contours, segment
from pyrpoc.rpoc.types import RPOCEditorState, RPOCImageInput, RPOCRoi


class _ImageViewer(QGraphicsView):
    def __init__(self, scene: QGraphicsScene, editor: "RPOCMaskEditor"):
        super().__init__(scene)
        self.editor = editor
        self.setRenderHints(self.renderHints() | QPainter.RenderHint.Antialiasing)
        self.setMouseTracking(True)
        self.setDragMode(QGraphicsView.DragMode.NoDrag)

        self.live_path_item: QGraphicsPathItem | None = None
        self.drawing = False
        self.current_path: QPainterPath | None = None
        self.current_points: list[QPointF] = []
        self.path_pen = QPen(Qt.GlobalColor.red, 2, Qt.PenStyle.SolidLine, Qt.PenCapStyle.RoundCap)

        self._zoom = 0
        self.roi_opacity = 0.4
        self.roi_items: dict[int, QGraphicsPathItem] = {}
        self.roi_label_items: dict[int, QGraphicsTextItem] = {}

        self.cursor_brush = QGraphicsEllipseItem(0, 0, 8, 8)
        self.cursor_brush.setBrush(QColor(100, 170, 255))
        self.cursor_brush.setPen(QPen(Qt.PenStyle.NoPen))
        self.cursor_brush.setZValue(1000)
        self.cursor_brush.setVisible(False)
        self.scene().addItem(self.cursor_brush)

    def wheelEvent(self, event):
        zoom_factor = 1.25 if event.angleDelta().y() > 0 else 0.8
        self._zoom += 1 if zoom_factor > 1 else -1
        self.scale(zoom_factor, zoom_factor)

    def mousePressEvent(self, event):
        if event.button() == Qt.MouseButton.RightButton:
            pos = self.mapToScene(event.pos())
            if not self.drawing:
                self.drawing = True
                self.current_points = [pos]
                self.current_path = QPainterPath()
                self.current_path.moveTo(pos)
                if self.live_path_item:
                    self.scene().removeItem(self.live_path_item)
                self.live_path_item = self.scene().addPath(self.current_path, self.path_pen)
            else:
                self._finish_path(pos)
            return

        if event.button() == Qt.MouseButton.LeftButton and not self.drawing:
            self.setDragMode(QGraphicsView.DragMode.ScrollHandDrag)
            self.viewport().setCursor(Qt.CursorShape.ClosedHandCursor)
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if self.drawing and self.current_path is not None:
            pos = self.mapToScene(event.pos())
            self.current_points.append(pos)
            self.current_path.lineTo(pos)
            if self.live_path_item:
                self.live_path_item.setPath(self.current_path)
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.MouseButton.LeftButton and not self.drawing:
            self.setDragMode(QGraphicsView.DragMode.NoDrag)
            self.viewport().setCursor(Qt.CursorShape.ArrowCursor)
        super().mouseReleaseEvent(event)

    def _finish_path(self, end_pos: QPointF) -> None:
        if self.current_path is None:
            return
        self.drawing = False
        self.current_points.append(end_pos)
        self.current_path.lineTo(end_pos)
        self.current_path.closeSubpath()
        if self.live_path_item:
            self.scene().removeItem(self.live_path_item)
            self.live_path_item = None

        points = [(float(p.x()), float(p.y())) for p in self.current_points]
        self.editor.add_roi(points)
        self.current_path = None
        self.current_points = []

    def add_roi_graphics(self, roi_id: int, points: list[tuple[float, float]]) -> None:
        if roi_id in self.roi_items:
            self.remove_roi_graphics(roi_id)
        if len(points) < 3:
            return

        path = QPainterPath()
        path.moveTo(QPointF(points[0][0], points[0][1]))
        for x, y in points[1:]:
            path.lineTo(QPointF(x, y))
        path.closeSubpath()

        color = self._random_color()
        roi_item = self.scene().addPath(path, QPen(color, 2))
        roi_item.setBrush(color)
        roi_item.setOpacity(self.roi_opacity if self.editor.state.show_rois else 0.0)
        self.roi_items[roi_id] = roi_item

        lbl = self._create_roi_label(roi_id, points)
        self.roi_label_items[roi_id] = lbl

    def remove_roi_graphics(self, roi_id: int) -> None:
        roi_item = self.roi_items.pop(roi_id, None)
        if roi_item is not None:
            self.scene().removeItem(roi_item)
        lbl = self.roi_label_items.pop(roi_id, None)
        if lbl is not None:
            self.scene().removeItem(lbl)

    def clear_roi_graphics(self) -> None:
        for roi_id in list(self.roi_items.keys()):
            self.remove_roi_graphics(roi_id)

    def _create_roi_label(self, roi_id: int, points: list[tuple[float, float]]) -> QGraphicsTextItem:
        xs = [p[0] for p in points]
        ys = [p[1] for p in points]
        cx, cy = sum(xs) / len(xs), sum(ys) / len(ys)
        label_item = QGraphicsTextItem(str(roi_id))
        label_item.setFont(QFont("Arial", 14, QFont.Weight.Bold))
        label_item.setDefaultTextColor(Qt.GlobalColor.white)
        bounds = label_item.boundingRect()
        label_item.setPos(cx - bounds.width() / 2, cy - bounds.height() / 2)
        label_item.setZValue(999)
        label_item.setVisible(self.editor.state.show_rois and self.editor.state.show_labels)
        self.scene().addItem(label_item)
        return label_item

    def _random_color(self) -> QColor:
        return QColor(random.randint(50, 255), random.randint(50, 255), random.randint(50, 255))

    def update_roi_visibility(self) -> None:
        for roi_item in self.roi_items.values():
            roi_item.setOpacity(self.roi_opacity if self.editor.state.show_rois else 0.0)
        for item in self.roi_label_items.values():
            item.setVisible(self.editor.state.show_rois and self.editor.state.show_labels)


class RPOCMaskEditor(QMainWindow):
    mask_created = pyqtSignal(object)

    _SLIDER_MIN = 0
    _SLIDER_MAX = 10000

    def __init__(
        self,
        image_data: np.ndarray | RPOCImageInput | None = None,
        parent: QWidget | None = None,
    ):
        super().__init__(parent)
        self.setWindowTitle("RPOC Mask Editor")
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose, True)

        self.state = RPOCEditorState()
        self._emitted = False
        self._closing_via_create = False
        self.threshold_mode = "normalized_0_1"
        self.data_min = 0.0
        self.data_max = 1.0

        self.roi_table = QTableWidget(0, 5)
        self.roi_table.setHorizontalHeaderLabels(
            ["ROI Name", "Coordinates", "Lower Threshold", "Upper Threshold", "Modulation Level"]
        )
        self.roi_table.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.roi_table.customContextMenuRequested.connect(self._show_table_context_menu)

        load_button = QPushButton("Load Image")
        load_button.clicked.connect(self._load_image)
        segment_button = QPushButton("Segment")
        segment_button.clicked.connect(self._segment)
        preview_button = QPushButton("Preview")
        preview_button.clicked.connect(self._preview_mask)
        save_button = QPushButton("Save Mask")
        save_button.clicked.connect(self._save_mask)
        create_button = QPushButton("Create Mask")
        create_button.clicked.connect(self.create_and_close)

        self.mask_checkbox = QCheckBox("Mask [M]")
        self.mask_checkbox.setChecked(True)
        self.mask_checkbox.stateChanged.connect(self._toggle_mask_visibility)
        self.label_checkbox = QCheckBox("Labels [N]")
        self.label_checkbox.setChecked(True)
        self.label_checkbox.stateChanged.connect(self._toggle_label_visibility)

        self.threshold_slider = QRangeSlider(Qt.Orientation.Horizontal)
        self.threshold_slider.setMinimum(self._SLIDER_MIN)
        self.threshold_slider.setMaximum(self._SLIDER_MAX)
        self.threshold_slider.setValue((2000, 8000))
        self.threshold_slider.valueChanged.connect(self._on_threshold_slider_changed)

        top_row = QHBoxLayout()
        top_row.addWidget(load_button)
        top_row.addWidget(segment_button)
        top_row.addWidget(preview_button)
        top_row.addWidget(save_button)
        top_row.addWidget(create_button)
        top_row.addStretch(1)

        toggle_row = QHBoxLayout()
        toggle_row.addWidget(self.mask_checkbox)
        toggle_row.addWidget(self.label_checkbox)
        toggle_row.addWidget(QLabel("Threshold"))
        toggle_row.addWidget(self.threshold_slider, 1)

        self.channels_row = QHBoxLayout()
        self.channel_checkboxes: list[QCheckBox] = []
        self.image_colors = [
            (255, 0, 0),
            (0, 255, 0),
            (0, 0, 255),
            (255, 255, 0),
            (0, 255, 255),
            (255, 0, 255),
        ]

        self.image_scene = QGraphicsScene()
        self.image_item = self.image_scene.addPixmap(QPixmap())
        self.image_view = _ImageViewer(self.image_scene, self)

        left = QVBoxLayout()
        left.addLayout(self.channels_row)
        left.addLayout(top_row)
        left.addLayout(toggle_row)
        left.addWidget(self.image_view, 1)

        root = QHBoxLayout()
        root.addLayout(left, 3)
        root.addWidget(self.roi_table, 2)

        central = QWidget(self)
        central.setLayout(root)
        self.setCentralWidget(central)

        if image_data is not None:
            self.set_image_data(image_data)
        else:
            self._update_displayed_image()

    def set_image_data(self, image_data: np.ndarray | RPOCImageInput) -> None:
        if isinstance(image_data, RPOCImageInput):
            image_input = image_data
        else:
            image_input = RPOCImageInput.from_array(image_data)

        self.state.image_input = image_input
        self.state.channel_visibility = [True] * image_input.data.shape[0]
        self.state.rois = []
        self.state.next_roi_id = 1
        self.state.show_rois = True
        self.state.show_labels = True
        self.mask_checkbox.setChecked(True)
        self.label_checkbox.setChecked(True)

        self.roi_table.setRowCount(0)
        self.image_view.clear_roi_graphics()
        self._clear_channel_widgets()
        self._build_channel_widgets()
        self._adjust_threshold_slider()
        self._update_displayed_image()

    def _clear_channel_widgets(self) -> None:
        while self.channels_row.count():
            item = self.channels_row.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()
        self.channel_checkboxes.clear()

    def _build_channel_widgets(self) -> None:
        image_input = self.state.image_input
        if image_input is None:
            self.channels_row.addStretch(1)
            return
        for i, label in enumerate(image_input.channel_labels):
            cb = QCheckBox(f"{label} [{i + 1}]")
            cb.setChecked(self.state.channel_visibility[i])
            cb.stateChanged.connect(lambda state, idx=i: self._on_channel_toggle(idx, state))
            self.channel_checkboxes.append(cb)
            self.channels_row.addWidget(cb)
        self.channels_row.addStretch(1)

    def _on_channel_toggle(self, idx: int, state: int) -> None:
        if idx < 0 or idx >= len(self.state.channel_visibility):
            return
        self.state.channel_visibility[idx] = bool(state)
        self._update_displayed_image()

    def _adjust_threshold_slider(self) -> None:
        image_input = self.state.image_input
        if image_input is None:
            return
        global_min = float(np.min(image_input.data))
        global_max = float(np.max(image_input.data))
        self._set_threshold_domain(global_min, global_max)

        span = max(self.data_max - self.data_min, 1e-9)
        low_data = self.data_min + 0.2 * span
        high_data = self.data_min + 0.8 * span
        if self.threshold_mode == "normalized_0_1":
            low_data = max(0.0, min(1.0, low_data))
            high_data = max(0.0, min(1.0, high_data))
            if high_data - low_data < 0.1:
                high_data = min(1.0, low_data + 0.5)

        low_i, high_i = self.data_to_slider(low_data, high_data)
        self.threshold_slider.blockSignals(True)
        self.threshold_slider.setValue((low_i, high_i))
        self.threshold_slider.blockSignals(False)

    def _set_threshold_domain(self, data_min: float, data_max: float) -> None:
        self.data_min = float(data_min)
        self.data_max = float(data_max)
        if self.data_max <= self.data_min:
            self.data_max = self.data_min + 1e-9
        if self.data_min >= 0.0 and self.data_max <= 1.0:
            self.threshold_mode = "normalized_0_1"
        else:
            self.threshold_mode = "raw_range"

    def _coerce_slider_bounds(self, low_i: int, high_i: int) -> tuple[int, int]:
        low = int(max(self._SLIDER_MIN, min(self._SLIDER_MAX, low_i)))
        high = int(max(self._SLIDER_MIN, min(self._SLIDER_MAX, high_i)))
        if low > high:
            low, high = high, low
        if low == high:
            if high < self._SLIDER_MAX:
                high += 1
            elif low > self._SLIDER_MIN:
                low -= 1
        return low, high

    def slider_to_data(self, low_i: int, high_i: int) -> tuple[float, float]:
        low_i, high_i = self._coerce_slider_bounds(low_i, high_i)
        if self.threshold_mode == "normalized_0_1":
            low = low_i / float(self._SLIDER_MAX)
            high = high_i / float(self._SLIDER_MAX)
        else:
            frac_low = low_i / float(self._SLIDER_MAX)
            frac_high = high_i / float(self._SLIDER_MAX)
            span = self.data_max - self.data_min
            low = self.data_min + frac_low * span
            high = self.data_min + frac_high * span
        if high < low:
            low, high = high, low
        return float(low), float(high)

    def data_to_slider(self, low: float, high: float) -> tuple[int, int]:
        low_f = float(low)
        high_f = float(high)
        if high_f < low_f:
            low_f, high_f = high_f, low_f
        if self.threshold_mode == "normalized_0_1":
            low_i = int(round(max(0.0, min(1.0, low_f)) * self._SLIDER_MAX))
            high_i = int(round(max(0.0, min(1.0, high_f)) * self._SLIDER_MAX))
        else:
            span = max(self.data_max - self.data_min, 1e-9)
            low_i = int(round((low_f - self.data_min) / span * self._SLIDER_MAX))
            high_i = int(round((high_f - self.data_min) / span * self._SLIDER_MAX))
        return self._coerce_slider_bounds(low_i, high_i)

    def _get_threshold_bounds(self) -> tuple[float, float]:
        low_i, high_i = self.threshold_slider.value()
        return self.slider_to_data(int(low_i), int(high_i))

    def _on_threshold_slider_changed(self, _value: tuple[int, int]) -> None:
        self._update_displayed_image()

    def _update_displayed_image(self) -> None:
        image_input = self.state.image_input
        if image_input is None:
            self.image_item.setPixmap(QPixmap())
            return
        data = image_input.data
        _, h, w = data.shape
        rgb_overlay = np.zeros((h, w, 3), dtype=np.uint8)
        low, high = self._get_threshold_bounds()
        high = max(high, low + 1e-9)

        for i, img in enumerate(data):
            visible = i < len(self.state.channel_visibility) and self.state.channel_visibility[i]
            if not visible:
                continue
            color = np.array(self.image_colors[i % len(self.image_colors)], dtype=np.float32)
            mask = (img >= low) & (img <= high)
            norm = np.zeros_like(img, dtype=np.float32)
            if np.any(mask):
                clipped = np.clip(img.astype(np.float32), low, high)
                norm[mask] = (clipped[mask] - low) / (high - low)
            channel = (norm[..., None] * color).astype(np.uint8)
            rgb_overlay = np.clip(rgb_overlay + channel, 0, 255)

        qimg = QImage(rgb_overlay.data, w, h, 3 * w, QImage.Format.Format_RGB888)
        self.image_item.setPixmap(QPixmap.fromImage(qimg.copy()))
        self.image_scene.setSceneRect(QRectF(qimg.rect()))

    def _find_roi(self, roi_id: int) -> RPOCRoi | None:
        for roi in self.state.rois:
            if roi.roi_id == roi_id:
                return roi
        return None

    def _find_row_by_roi_id(self, roi_id: int) -> int:
        for row in range(self.roi_table.rowCount()):
            item = self.roi_table.item(row, 0)
            if item is None:
                continue
            rid = item.data(Qt.ItemDataRole.UserRole)
            if isinstance(rid, int) and rid == roi_id:
                return row
        return -1

    def _upsert_roi_row(self, roi: RPOCRoi) -> None:
        row = self._find_row_by_roi_id(roi.roi_id)
        if row < 0:
            row = self.roi_table.rowCount()
            self.roi_table.insertRow(row)

        item = QTableWidgetItem(f"ROI {roi.roi_id}")
        item.setData(Qt.ItemDataRole.UserRole, roi.roi_id)
        self.roi_table.setItem(row, 0, item)
        self.roi_table.setItem(
            row,
            1,
            QTableWidgetItem(", ".join(f"({x:.1f}, {y:.1f})" for x, y in roi.points)),
        )
        self.roi_table.setItem(row, 2, QTableWidgetItem(f"{roi.threshold_low:.6f}"))
        self.roi_table.setItem(row, 3, QTableWidgetItem(f"{roi.threshold_high:.6f}"))
        self.roi_table.setItem(row, 4, QTableWidgetItem(f"{roi.modulation_level:.6f}"))

    def add_roi(self, points: list[tuple[float, float]]) -> None:
        if len(points) < 3:
            return
        low, high = self._get_threshold_bounds()
        roi_id = self.state.next_roi_id
        self.state.next_roi_id += 1
        roi = RPOCRoi(
            roi_id=roi_id,
            points=[(float(x), float(y)) for x, y in points],
            threshold_low=low,
            threshold_high=high,
            modulation_level=0.5,
            active_channels=self.state.channel_visibility.copy(),
        )
        self.state.rois.append(roi)
        self.image_view.add_roi_graphics(roi.roi_id, roi.points)
        self._upsert_roi_row(roi)

    def _show_table_context_menu(self, pos):
        row = self.roi_table.indexAt(pos).row()
        if row < 0:
            return
        menu = QMenu(self)
        delete_action = menu.addAction("Delete ROI")
        action = menu.exec(self.roi_table.mapToGlobal(pos))
        if action == delete_action:
            self._delete_roi_row(row)

    def _delete_roi_row(self, row: int) -> None:
        idx_item = self.roi_table.item(row, 0)
        if idx_item is None:
            return
        roi_id = idx_item.data(Qt.ItemDataRole.UserRole)
        if not isinstance(roi_id, int):
            return
        self.roi_table.removeRow(row)
        self.state.rois = [roi for roi in self.state.rois if roi.roi_id != roi_id]
        self.image_view.remove_roi_graphics(roi_id)

    def _toggle_mask_visibility(self, state: int) -> None:
        self.state.show_rois = bool(state)
        self.image_view.update_roi_visibility()

    def _toggle_label_visibility(self, state: int) -> None:
        self.state.show_labels = bool(state)
        self.image_view.update_roi_visibility()

    def _get_composite(self) -> np.ndarray | None:
        image_input = self.state.image_input
        if image_input is None:
            return None
        visible = [
            img for i, img in enumerate(image_input.data) if i < len(self.state.channel_visibility) and self.state.channel_visibility[i]
        ]
        if not visible:
            return None
        composite = np.mean(visible, axis=0).astype(np.float32)
        max_val = float(np.max(composite)) if composite.size else 0.0
        if max_val > 1.0:
            composite = composite / max_val
        return np.clip(composite, 0, 1)

    def _segment(self) -> None:
        comp = self._get_composite()
        if comp is None:
            return
        low, high = self._get_threshold_bounds()
        if self.threshold_mode == "raw_range":
            span = max(self.data_max - self.data_min, 1e-9)
            low = (low - self.data_min) / span
            high = (high - self.data_min) / span
        low = float(max(0.0, min(1.0, low)))
        high = float(max(0.0, min(1.0, high)))

        labels = segment(comp, "threshold + components", low=low, high=high, min_area=120)
        contours = labels_to_contours(labels, min_area=120)
        if not contours:
            return
        for contour in contours:
            points = [(float(x), float(y)) for x, y in contour]
            self.add_roi(points)

    def _sync_roi_values_from_table(self) -> None:
        for row in range(self.roi_table.rowCount()):
            idx_item = self.roi_table.item(row, 0)
            if idx_item is None:
                continue
            roi_id = idx_item.data(Qt.ItemDataRole.UserRole)
            if not isinstance(roi_id, int):
                continue
            roi = self._find_roi(roi_id)
            if roi is None:
                continue
            try:
                low_val = float(self.roi_table.item(row, 2).text())
                high_val = float(self.roi_table.item(row, 3).text())
                mod_val = float(self.roi_table.item(row, 4).text())
            except Exception:
                continue
            if high_val < low_val:
                low_val, high_val = high_val, low_val
            roi.threshold_low = float(low_val)
            roi.threshold_high = float(high_val)
            roi.modulation_level = float(np.clip(mod_val, 0.0, 1.0))

    def generate_final_mask(self) -> np.ndarray | None:
        image_input = self.state.image_input
        if image_input is None:
            return None
        self._sync_roi_values_from_table()
        _, h, w = image_input.data.shape
        final_mask = np.zeros((h, w), dtype=np.uint8)

        for roi in self.state.rois:
            pts = np.array(
                [[int(round(x)), int(round(y))] for x, y in roi.points],
                dtype=np.int32,
            ).reshape(-1, 1, 2)
            if pts.shape[0] < 3:
                continue
            roi_mask = np.zeros((h, w), dtype=np.uint8)
            cv2.fillPoly(roi_mask, [pts], 255)

            combined = np.zeros((h, w), dtype=bool)
            for img, active in zip(image_input.data, roi.active_channels):
                if not active:
                    continue
                combined |= (img >= roi.threshold_low) & (img <= roi.threshold_high)

            final_mask[combined & (roi_mask == 255)] = int(np.clip(roi.modulation_level, 0.0, 1.0) * 255)
        return final_mask

    def _preview_mask(self) -> None:
        mask = self.generate_final_mask()
        if mask is None:
            return
        h, w = mask.shape
        qimg = QImage(mask.data, w, h, w, QImage.Format.Format_Grayscale8)
        dlg = QDialog(self)
        dlg.setWindowTitle("Mask Preview")
        layout = QVBoxLayout(dlg)
        lbl = QLabel()
        lbl.setPixmap(QPixmap.fromImage(qimg.copy()))
        layout.addWidget(lbl)
        dlg.resize(w, h)
        dlg.exec()

    def _save_mask(self) -> None:
        mask = self.generate_final_mask()
        if mask is None:
            return
        path, _ = QFileDialog.getSaveFileName(
            self,
            "Save Mask",
            "",
            "PNG (*.png);;TIFF (*.tif *.tiff);;All Files (*)",
        )
        if not path:
            return
        cv2.imwrite(path, mask)

    def _load_image(self) -> None:
        path, _ = QFileDialog.getOpenFileName(
            self,
            "Open Image",
            "",
            "Images (*.png *.jpg *.jpeg *.bmp *.tif *.tiff)",
        )
        if not path:
            return
        img = cv2.imread(path, cv2.IMREAD_UNCHANGED)
        if img is None:
            return
        if img.ndim == 2:
            data = img.astype(np.float32)
        elif img.ndim == 3:
            rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            data = np.moveaxis(rgb, -1, 0).astype(np.float32)
        else:
            return
        self.set_image_data(data)

    def create_and_close(self) -> None:
        if self._closing_via_create:
            return
        self._closing_via_create = True
        mask = self.generate_final_mask()
        if mask is not None:
            self._emitted = True
            self.mask_created.emit(mask.astype(np.uint8))
        self.close()

    def closeEvent(self, event):
        if not self._emitted and not self._closing_via_create:
            mask = self.generate_final_mask()
            if mask is not None and np.any(mask):
                self._emitted = True
                self.mask_created.emit(mask.astype(np.uint8))
        super().closeEvent(event)
